#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#ifdef HAVE_IMLIB
# include <Imlib.h>
#endif
#ifdef HAVE_IMLIB2
# include <X11/Xlib.h>
# include <Imlib2.h>
#endif
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include "imgio.h"
#include "guiutils.h"
#include "cdialog.h"
#include "fprompt.h"
#include "csd.h"
#include "fb.h"
#include "progressdialog.h"
#include "imgview.h"
#include "ivtitle.h"
#include "saveopts.h"
#include "rootwin.h"
#include "grab.h"
#include "grabopts.h"
#include "printopts.h"
#include "print.h"
#include "config.h"

#include "images/icon_camera_20x20.xpm"
#include "images/icon_open_20x20.xpm"
#include "images/icon_save_as_20x20.xpm"
#include "images/icon_info_20x20.xpm"
#include "images/icon_desktop_20x20.xpm"
#include "images/icon_print_20x20.xpm"

#include "icon_iv_20x20.xpm"


/* Define this to create the image viewer as a stand alone toplevel
 * window. Normally you do want to define this, since undefining it
 * will have some issues in do_new_viewer() (read the comments in that
 * function).
 */
#define CREATE_IMGVIEW_AS_TOPLEVEL      1


#ifdef HAVE_IMLIB
/* Handle used for passing to all Imlib functions, not needed for
 * Imlib2.
 */
void *imlib_handle = NULL;
#endif	/* HAVE_IMLIB */


/* Has changes string, displayed on title. */
#define HAS_CHANGES_STR	"(*)"

/* No comments string. */
#define NO_COMMENTS_STR	"None"


/* Image info structure. */
typedef struct {

	gchar *filename;	/* Can be NULL if never been saved. */
	gint width, height;	/* In pixels. */
	glong filesize;		/* Size in bytes on disk, can be 0 if
				 * not been saved yet.
				 */
	gboolean has_alpha;	/* TRUE if alpha channel exists. */
	gchar *creator;		/* Program that generated image, can be NULL. */
	gchar *author;		/* Author and/or e-mail, can be NULL. */
        gchar *comments;	/* Can be NULL. */
	gulong time_modified;	/* Last modified time in systime seconds. */

} iv_info_struct;

/* Viewer structure. */
typedef struct {

	/* Pointer back to core. */
	gpointer core_ptr;

	/* Marks if this viewer is busy processing. */
	gboolean processing;

	/* Image viewer. */
	imgview_struct *iv;

	/* Root window submenu. */
	GtkWidget *root_window_submenu;

	/* Important menu items. */
	GtkWidget	*set_bg_mi,
			*open_mi,
			*save_as_mi,
			*information_mi,
			*root_window_mi,
			*clear_mi,
			*grab_mi,
			*print_mi,
			*new_viewer_mi;

	gboolean has_changes;	/* TRUE if image has modifications that
				 * have not been saved yet.
				 */

	/* Image info structure. */
	iv_info_struct info;

	/* Background color, 4 bytes in RGBA format. */
	guint8 bg_color[4];

} iv_viewer_struct;

/* Core structure. */
typedef struct {

	/* Image viewers. */
	iv_viewer_struct **viewer;
	gint total_viewers;

	/* Untitled image counter, increments when a new image is created
	 * or obtained from grab.
	 */
	gint untitled_count;

	/* Grab cursor. */
	GdkCursor *grab_cur;

	/* GtkWindow used for grab window operations, pointer events are
	 * specified to be relative to this window.
	 */
	GtkWidget *grab_window;

	/* Options. */
	gboolean show_tooltips;

	/* These options reflect their respective value set from the
	 * command line or default. They should neither be modified nor
	 * accessed as up to date values during run time.
	 *
	 * To get the up to date value, check the viewer's iv for the
	 * actual curent value.
	 */
        gboolean show_toolbar;
        gboolean show_values;
	gboolean show_statusbar;
        gboolean show_image_on_wm_icon;
        gint quality;
	/* Geometry setting at startup, this is only defined if member
	 * geometry_uset is TRUE.
	 */
	gboolean geometry_uset;
        GdkRectangle geometry;
	/* Resize image viewer whenever a new image is loaded. */
	gboolean resize_on_load;

        /* Imlib fallback code:
         *
         *      0       No fallbacks
         *      1       Use ImageMagick & NETPBM
         */
	gint imlib_fallback;

	/* File format save options. */
	gboolean gif_interlaced;
	gfloat jpeg_quality;		/* 0.0 to 1.0. */
	gboolean jpeg_color;
	gint png_compression;		/* 0 to 9. */
	gint png_format;		/* 0 = B&W
					 * 1 = Greyscale
					 * 2 = Greyscale Alpha
					 * 3 = RGB
					 * 4 = RGBA
					 */
	gboolean png_interlaced;
	gint tga_format;		/* 0 = Greyscale
					 * 1 = RGB
					 * 2 = RGBA
					 */

	/* Grab window options. */
	glong grab_delay;		/* In milliseconds. */
	gboolean grab_include_decorations;
	gboolean grab_hide_windows;

	/* Print options. */
	print_values_struct print_values;
	gchar *print_command;		/* Last print command. */
	gchar *print_file;		/* Last print file name. */

} iv_core_struct;


/* Structure used as callback data for ImgIO*() functions. */
typedef struct {

	GtkWidget *ref_toplevel;
	iv_viewer_struct *viewer;

	gchar *path;

} iv_imgio_cb_struct;


/* Local global current directory. */
static gchar current_directory[PATH_MAX];


static gint delete_event(
        GtkWidget *widget, GdkEvent *event, gpointer data
);
static void iv_dnd_recieved_cb(
        GtkWidget *widget, GdkDragContext *dc, gint x, gint y,
        GtkSelectionData *selection_data, guint info, guint t,
        gpointer data
);
static iv_viewer_struct *do_new_viewer(
	iv_core_struct *core_ptr,
	iv_viewer_struct ***v, gint *total, gint *n,
        gboolean show_toolbar,
        gboolean show_values,
	gboolean show_statusbar,
        gboolean show_image_on_wm_icon,
	gint quality,
	GdkRectangle *geometry,
	gboolean load_title
);
static void viewer_delete(iv_viewer_struct *v);
static void grab_window_destroy(iv_core_struct *core_ptr);
static void iv_update_image_info(
	iv_viewer_struct *v,
	const gchar *path,		/* Can be NULL. */
	struct stat *stat_buf,		/* Can be NULL. */
	gint width, gint height,
	gint bpl, gint bpp,
	gboolean has_alpha,
	const gchar *creator,		/* Can be NULL. */
	const gchar *author,		/* Can be NULL. */
	const gchar *comments,		/* Can be NULL. */
	gulong time_modified		/* In systime seconds. */
);
static void do_load_image(
        iv_viewer_struct *v, const gchar *path, gboolean resize_on_load
);
static void do_save_image(
        iv_viewer_struct *iv, const gchar *path, const gchar *ext
);
static void do_load_image_to_root_centered(const gchar *path);
static void do_load_image_to_root_tofit(const gchar *path);
static void do_load_image_to_root_tocover(const gchar *path);
static void do_load_image_to_root_tiled(const gchar *path);

static void imgview_changed_cb(
        gpointer imgview, imgview_image_struct *image, gpointer data
);
static void set_bg_cb(GtkWidget *widget, gpointer data);
static void grab_start_cb(GtkWidget *widget, gpointer data);
static gint iv_load_progress_cb(
	gpointer data,			/* Client data. */
	gint min, gint max,		/* Progress. */
	gint width, gint height,
	gint bpl, gint bpp,
	guint8 *rgba_data
);
static gint iv_save_progress_cb(
        gpointer data,                  /* Client data. */
        gint min, gint max,             /* Progress. */
        gint width, gint height,
        gint bpl, gint bpp,
        guint8 *rgba_data
);
static void open_image_cb(GtkWidget *widget, gpointer data);
static void save_image_cb(GtkWidget *widget, gpointer data);
static void display_on_root_centered_cb(GtkWidget *widget, gpointer data);
static void display_on_root_tofit_cb(GtkWidget *widget, gpointer data);
static void display_on_root_tocover_cb(GtkWidget *widget, gpointer data);
static void display_on_root_tiled_cb(GtkWidget *widget, gpointer data);
static void clear_image_cb(GtkWidget *widget, gpointer data);
static void information_cb(GtkWidget *widget, gpointer data);
static void print_cb(GtkWidget *widget, gpointer data);

static void new_viewer(GtkWidget *widget, gpointer data);
static void update_view_title(iv_viewer_struct *v, const gchar *new_path);
static void update_view_menus(iv_viewer_struct *v);
static void clear_info(iv_info_struct *info);

static gboolean extcmp(const gchar *s, const gchar *ext);
static void get_args_value(
        char *arg, char **parm_rtn, char **val_rtn
);
static gboolean stristrue(const gchar *s);


#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)	(((x) < 0) ? ((x) * -1) : (x))


/*
 *	Called when the image viewer toplevel has closed.
 */
static gint delete_event(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gboolean need_quit_main = FALSE;
	gint i, status;
	GtkWidget *w;
	iv_viewer_struct *v = (iv_viewer_struct *)data;
	iv_core_struct *core_ptr;
        if(v == NULL)
            return(TRUE);

	core_ptr = (iv_core_struct *)v->core_ptr;
	if(core_ptr == NULL)
	    return(TRUE);


	/* Check if any dialogs are still querying user. */
        if(CDialogIsQuery())
            return(TRUE);

        if(CSDIsQuery())
            return(TRUE);

        if(FileBrowserIsQuery())
            return(TRUE);

	if(ProgressDialogIsQuery())
	    return(TRUE);

	/* Skip if this viewer is busy processing. */
	if(v->processing)
	    return(TRUE);


	/* Query user if there are unsaved changes. */
	if(v->has_changes)
	{
	    /* Get image viewer toplevel widget if it is a GtkWindow. */
	    if(ImgViewToplevelIsWindow(v->iv))
		w = ImgViewGetToplevelWidget(v->iv);
	    else
		w = NULL;

            CDialogSetTransientFor(w);
            status = CDialogGetResponse(
"Save Changes?",
"The image data contains changes that have not been\n\
saved, do you want to save those changes?",
"There are changes made to the image data on this\n\
Image Viewer that have not been saved. If you want\n\
to save those changes then click on `Yes'. If you\n\
want to discard those changes then click on `No'.\n\
If you decide to not close the Image Viewer then\n\
click on `Cancel'.",
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO |
                CDIALOG_BTNFLAG_CANCEL | CDIALOG_BTNFLAG_HELP,
                CDIALOG_BTNFLAG_YES
            );
            CDialogSetTransientFor(NULL);
            switch(status)
            {
	      case CDIALOG_RESPONSE_CANCEL:
/*		reenterent = FALSE; */
		return(TRUE);
		break;

	      case CDIALOG_RESPONSE_NO:
                /* Discard changes. */
                break;

              default:
                /* Call save image callback. */
                save_image_cb(v->save_as_mi, v);
                /* Check if save failed by checking if has_changes is
                 * still TRUE.
                 */
                if(v->has_changes)
                {
/*		    reenterent = FALSE; */
                    return(TRUE);
                }
                break;
            }
	}


	/* Begin deleting this viewer and deallocating its resources. */

	/* Remove this viewer pointer from core's viewers list. */
	for(i = 0; i < core_ptr->total_viewers; i++)
	{
	    if(core_ptr->viewer[i] == v)
		core_ptr->viewer[i] = NULL;
	}

	/* Delete this viewer. */
	viewer_delete(v);
	v = NULL;

	/* Check if there are no viewers left on the core structure. */
	for(i = 0; i < core_ptr->total_viewers; i++)
        {
            if(core_ptr->viewer[i] != NULL)
		break;
        }
	/* No viewers left? */
	if(i >= core_ptr->total_viewers)
	{
	    need_quit_main = TRUE;
	}

	/* Need to break out of the toplevel GTK+ main loop? */
	if(need_quit_main)
	    gtk_main_quit();

        return(TRUE);
}

/*
 *	DND data recieved callback for view drawable widget.
 */
static void iv_dnd_recieved_cb(
	GtkWidget *widget, GdkDragContext *dc, gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
        gpointer data
)
{
	GtkWidget *w;
	imgview_struct *iv;
	iv_core_struct *core_ptr;
	iv_viewer_struct *v = (iv_viewer_struct *)data;
        if((widget == NULL) || (v == NULL) || (dc == NULL))
	    return;

        if(v->processing)
            return;

        iv = v->iv;
        core_ptr = (iv_core_struct *)v->core_ptr;
        if((iv == NULL) || (core_ptr == NULL))
            return;

        /* Mark viewer as processing. */
	ImgViewSetBusy(iv, TRUE);
        v->processing = TRUE;
	

	/* Get image viewer's view widget. */
	w = (GtkWidget *)ImgViewGetViewWidget(iv);

	/* View widget the same as the callback widget? */
	if(w == widget)
	{
	    /* Check if the DND target type is one that we support. */
	    if((info == IV_DND_TYPE_INFO_TEXT_PLAIN) ||
               (info == IV_DND_TYPE_INFO_TEXT_URI_LIST) ||
	       (info == IV_DND_TYPE_INFO_STRING)
	    )
	    {
		const gchar *cstrptr;
		const gchar *url = (const gchar *)selection_data->data;
		gchar *full_path;


		/* The DND data should be a URL string, we need to parse
		 * it and get the full path portion of that URL string.
		 */
		cstrptr = strstr(url, "://");
		if(cstrptr != NULL)
		    cstrptr += strlen("://");
		else
		    cstrptr = url;

		cstrptr = strchr(cstrptr, '/');
		full_path = (cstrptr != NULL) ? g_strdup(cstrptr) : NULL;


		/* Got full path to object to be loaded? */
		if(full_path != NULL)
		{
		    do_load_image(v, full_path, core_ptr->resize_on_load);

		    g_free(full_path);
		    full_path = NULL;
		}
	    }
	}

	/* Mark viewer as no longer processing. */
	v->processing = FALSE;
        ImgViewSetBusy(iv, FALSE);

	/* Update menus. */
	update_view_menus(v);
}

/*
 *	Procedure to create a new image viewer or map an existing
 *	one that is initialized but unmapped.
 */
static iv_viewer_struct *do_new_viewer(
	iv_core_struct *core_ptr,
        iv_viewer_struct ***v, gint *total, gint *n,
        gboolean show_toolbar,
        gboolean show_values,
	gboolean show_statusbar,
        gboolean show_image_on_wm_icon,
	gint quality,
	GdkRectangle *geometry,
	gboolean load_title
)
{
        gint i;
        GtkWidget *w, *parent;
        GtkWidget *toplevel = NULL;
        GtkWidget *iv_toplevel, *menu;
        imgview_struct *iv = NULL;
	iv_viewer_struct *v_ptr = NULL;


	if(n != NULL)
	    (*n) = -1;

	if((v == NULL) || (total == NULL) || (core_ptr == NULL))
	    return(NULL);

	/* Check if an image viewer is available on the given list,
	 * it must be allocated, initialized, and unmapped.
	 */
	for(i = 0; i < (*total); i++)
	{
	    v_ptr = (*v)[i];
	    if(v_ptr == NULL)
		continue;

	    iv = v_ptr->iv;
	    if(iv == NULL)
		continue;

	    if(!iv->map_state)
		break;
	}
	/* Found an available viewer in the list? */
	if(i < (*total))
	{
	    /* Update viewer index return. */
	    if(n != NULL)
		(*n) = i;

	    /* Reset, load title image, and update menus. */
	    ImgViewReset(iv, FALSE);
	    if(load_title)
		iv_load_title(iv, core_ptr->resize_on_load);
	    update_view_menus(v_ptr);

	    /* Reset geometry as needed. */
	    if(ImgViewToplevelIsWindow(iv))
	    {
		w = ImgViewGetToplevelWidget(iv);
		if((geometry != NULL) && (w != NULL))
		{
		    gtk_widget_set_uposition(
			w, geometry->x, geometry->y
		    );
		    gtk_widget_set_usize(
			w, geometry->width, geometry->height
		    );
		}
	    }

	    /* Map it. */
	    ImgViewMap(iv);

	    /* Return pointer to existing viewer structure. */
	    return(v_ptr);
	}


	/* Allocate new viewer structure. */
	v_ptr = (iv_viewer_struct *)g_malloc0(sizeof(iv_viewer_struct));
	if(v_ptr == NULL)
	    return(NULL);

	/* Reset viewer structure values. */
	v_ptr->core_ptr = core_ptr;
	v_ptr->processing = FALSE;
	v_ptr->has_changes = FALSE;
	v_ptr->bg_color[0] = 0xff;
        v_ptr->bg_color[1] = 0xff;
        v_ptr->bg_color[2] = 0xff;
        v_ptr->bg_color[3] = 0xff;

	/* Reset image info structure. */
	clear_info(&v_ptr->info);

        /* Begin creating image viewer. */
#ifdef CREATE_IMGVIEW_AS_TOPLEVEL
        /* Create image viewer as a toplevel window. */
	v_ptr->iv = iv = ImgViewNew(
            show_toolbar,               /* Show tool bar. */
            show_values,                /* Show values. */
	    show_statusbar,		/* Show status bar. */
            show_image_on_wm_icon,      /* Show image on wm icon. */
	    quality,
            TRUE,                       /* Toplevel to be a GtkWindow? */
            &iv_toplevel
        );
        if(iv == NULL)
	{
	    g_free(v_ptr);
            return(NULL);
	}

        /* Get image viewer's toplevel widget. */
        parent = w = toplevel = iv_toplevel;

	/* Allow image viewer to crop. */
	ImgViewAllowCrop(iv, TRUE);

        /* Set image viewer changed callback. */
        ImgViewSetChangedCB(
            iv, imgview_changed_cb, v_ptr
        );

	/* Setup image viewer's toplevel window. */
	if(iv_toplevel != NULL)
	{
	    if(geometry != NULL)
	    {
		gtk_widget_set_uposition(
		    iv_toplevel, geometry->x, geometry->y
		);
		gtk_widget_set_usize(
		    iv_toplevel, geometry->width, geometry->height
		);
	    }
	    /* Set initial title. */
	    gtk_window_set_title(
		GTK_WINDOW(iv_toplevel),
		PROG_NAME " Version " PROG_VERSION
	    );
	}

        /* Connect "delete_event" signal for image viewer's toplevel. */
        gtk_signal_connect(
            GTK_OBJECT(w), "delete_event",
            GTK_SIGNAL_FUNC(delete_event), v_ptr
        );
#else
/* Note that the toplevel window will never be destroyed untill
 * program exits, this preprocessor condition is only for testing
 * purposes.
 */
        /* Create image viewer as a child widget. */
	v_ptr->iv = iv = ImgViewNew(
            show_toolbar,               /* Show tool bar. */
            show_values,                /* Show values. */
	    show_statusbar,		/* Show status bar. */
            show_image_on_wm_icon,      /* Show image on wm icon. */
	    quality,
            FALSE,                      /* Toplevel to be a GtkWindow? */
            &iv_toplevel
        );
        if(iv == NULL)
        {
            g_free(v_ptr);
            return(NULL);
        }

        /* Allow image viewer to crop. */
        ImgViewAllowCrop(iv, TRUE);

	/* Set image viewer changed callback. */
	ImgViewSetChangedCB(
	    iv, imgview_changed_cb, v_ptr
	);

        /* Create our toplevel window which will be the parent for
         * the image viewer widget.
         */
        gtk_widget_push_visual(gdk_rgb_get_visual());
        gtk_widget_push_colormap(gdk_rgb_get_cmap());
        toplevel = parent = w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
        gtk_widget_pop_visual();
        gtk_widget_pop_colormap();
        gtk_window_set_policy(
            GTK_WINDOW(w), TRUE, TRUE, TRUE
        );
        if((w != NULL) && (geometry != NULL))
	{
            gtk_widget_set_uposition(
                w, geometry->x, geometry->y
            );
            gtk_widget_set_usize(
		w, geometry->width, geometry->height
	    );
	}
	/* Set initial title. */
	gtk_window_set_title(
	    GTK_WINDOW(w),
	    PROG_NAME " Version " PROG_VERSION
	);
        gtk_signal_connect(
            GTK_OBJECT(w), "delete_event",
            GTK_SIGNAL_FUNC(delete_event), v_ptr
        );
        gtk_container_border_width(GTK_CONTAINER(w), 0);

        /* Place image viewer toplevel widget into our toplevel window. */
        gtk_container_add(GTK_CONTAINER(parent), iv_toplevel);
#endif

	/* Set up view widget for DND. */
	w = (GtkWidget *)ImgViewGetViewWidget(iv);
	if(w != NULL)
	{
	    const GtkTargetEntry dnd_tar_types[] = {
{"text/plain",		0,		IV_DND_TYPE_INFO_TEXT_PLAIN},
{"text/uri-list",	0,		IV_DND_TYPE_INFO_TEXT_URI_LIST},
{"STRING",		0,		IV_DND_TYPE_INFO_STRING}
	    };
	    GUIDNDSetTar(
		w,
		dnd_tar_types,
		sizeof(dnd_tar_types) / sizeof(GtkTargetEntry),
		GDK_ACTION_DEFAULT |		/* Drag actions. */
		GDK_ACTION_LINK |
		GDK_ACTION_MOVE |
		GDK_ACTION_COPY,
		GDK_ACTION_MOVE,		/* Def act if same. */
		GDK_ACTION_COPY,		/* Def act. */
		iv_dnd_recieved_cb,
		v_ptr
	    );
	}




        /* Get menu from image viewer. */
        menu = (GtkWidget *)ImgViewGetMenuWidget(iv);
        if(menu != NULL)
        {
            GtkWidget *fw, *submenu;
            gint accel_key;
            gpointer accel_group = ImgViewGetAccelGroup(iv);
            guint accel_mods;
            u_int8_t **icon;
            const gchar *label = NULL;
            gpointer mclient_data = v_ptr;
            void (*func_cb)(GtkWidget *w, gpointer) = NULL;

#define DO_ADD_MENU_ITEM_LABEL		\
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accel_group, \
  icon, label, accel_key, accel_mods, (gpointer *)&fw, \
  mclient_data, func_cb \
 ); \
}
#define DO_ADD_MENU_ITEM_SUBMENU	\
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SUBMENU, accel_group, \
  icon, label, accel_key, accel_mods, (gpointer *)&fw, \
  mclient_data, func_cb \
 ); \
 if(w != NULL) \
  GUIMenuItemSetSubMenu(w, submenu); \
}
#define DO_ADD_MENU_SEP         \
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); \
}

	    icon = NULL;
	    label = "Background Color...";
            accel_key = 'b';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = set_bg_cb;
            DO_ADD_MENU_ITEM_LABEL
            v_ptr->set_bg_mi = w;

	    DO_ADD_MENU_SEP

            icon = (u_int8_t **)icon_open_20x20_xpm;
            label = "Open...";
            accel_key = 'o';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = open_image_cb;
            DO_ADD_MENU_ITEM_LABEL
	    v_ptr->open_mi = w;

            icon = (u_int8_t **)icon_save_as_20x20_xpm;
            label = "Save As...";
            accel_key = 's';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = save_image_cb;
            DO_ADD_MENU_ITEM_LABEL
	    v_ptr->save_as_mi = w;

	    DO_ADD_MENU_SEP

            icon = (u_int8_t **)icon_info_20x20_xpm;
            label = "Information";
            accel_key = 'i';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = information_cb;
            DO_ADD_MENU_ITEM_LABEL
            v_ptr->information_mi = w;


	    /* Create display on root submenu. */
	    v_ptr->root_window_submenu = submenu = GUIMenuCreate();
	    if(submenu != NULL)
	    {
		GtkWidget *menu;	/* Overload. */

		menu = submenu;

                icon = NULL;
                label = "Centered";
                accel_key = 0;
                accel_mods = 0;
                func_cb = display_on_root_centered_cb;
                DO_ADD_MENU_ITEM_LABEL

                icon = NULL;
                label = "To Fit";
                accel_key = 0;
                accel_mods = 0;
		func_cb = display_on_root_tofit_cb;
		DO_ADD_MENU_ITEM_LABEL

                icon = NULL;
                label = "To Cover";
                accel_key = 0;
                accel_mods = 0;
                func_cb = display_on_root_tocover_cb;
                DO_ADD_MENU_ITEM_LABEL

                icon = NULL;
                label = "Tiled";
                accel_key = 0;
                accel_mods = 0;
                func_cb = display_on_root_tiled_cb;
                DO_ADD_MENU_ITEM_LABEL
	    }
            icon = (u_int8_t **)icon_desktop_20x20_xpm;
            label = "Display On Root";
            accel_key = 0;
            accel_mods = 0;
            func_cb = NULL;
            DO_ADD_MENU_ITEM_SUBMENU
	    v_ptr->root_window_mi = w;

            icon = NULL;
            label = "Clear";
            accel_key = 0;
            accel_mods = 0;
            func_cb = clear_image_cb;
            DO_ADD_MENU_ITEM_LABEL
	    v_ptr->clear_mi = w;

	    DO_ADD_MENU_SEP

            icon = (u_int8_t **)icon_camera_20x20_xpm;
            label = "Grab Window...";
            accel_key = 'g';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = grab_start_cb;
            DO_ADD_MENU_ITEM_LABEL
	    v_ptr->grab_mi = w;

            DO_ADD_MENU_SEP

            icon = (u_int8_t **)icon_print_20x20_xpm;
            label = "Print...";
            accel_key = 'p';
            accel_mods = GDK_CONTROL_MASK;
            func_cb = print_cb;
            DO_ADD_MENU_ITEM_LABEL
	    v_ptr->print_mi = w;

	    DO_ADD_MENU_SEP

            icon = (u_int8_t **)icon_iv_20x20_xpm;
            label = "New Image Viewer";
            accel_key = 0;
            accel_mods = 0;
            func_cb = new_viewer;
            DO_ADD_MENU_ITEM_LABEL
	    v_ptr->new_viewer_mi = w;

#undef DO_ADD_MENU_ITEM_LABEL
#undef DO_ADD_MENU_ITEM_SUBMENU
#undef DO_ADD_MENU_SEP
        }

        /* Record newly created image viewer on core. */
	/* Sanitize total. */
	if((*total) < 0)
	    (*total) = 0;
	/* Look for available NULL pointer in the given list. */
        for(i = 0; i < (*total); i++)
        {
            if((*v)[i] == NULL)
                break;
	}
	if(i < (*total))
	{
	    /* Got available NULL pointer at index i. */
	    (*v)[i] = v_ptr;
	}
	else
	{
	    /* Need to allocate one more pointer. */
	    i = (*total);
	    (*total) = i + 1;
	    (*v) = (iv_viewer_struct **)g_realloc(
		*v, (*total) * sizeof(iv_viewer_struct *)
	    );
	    if((*v) != NULL)
		(*v)[i] = v_ptr;
	    else
		(*total) = 0;
	}

	/* Update viewer index return. */
	if(n != NULL)
	    (*n) = i;


        /* Load title image? */
	if(load_title)
	    iv_load_title(iv, core_ptr->resize_on_load);

        /* Update menus. */
        update_view_menus(v_ptr);

        /* Make image viewer toplevel widget eligable for mapping. */
        ImgViewMap(iv);

#ifndef CREATE_IMGVIEW_AS_TOPLEVEL
        /* Map our toplevel window. */
        gtk_widget_show(toplevel);
#endif

	return(v_ptr);
}

/*
 *	Deallocates all resources on the given viewer structure including
 *	the structure itself.
 */
static void viewer_delete(iv_viewer_struct *v)
{
	imgview_struct *iv;
	iv_core_struct *core_ptr;

	if(v == NULL)
	    return;

	iv = v->iv;
	core_ptr = (iv_core_struct *)v->core_ptr;

	/* Begin destroying widgets that we created on the image viewer. */
	v->set_bg_mi = NULL;
	v->open_mi = NULL;
	v->save_as_mi = NULL;
	v->information_mi = NULL;
	v->root_window_mi = NULL;
	v->clear_mi = NULL;
	v->grab_mi = NULL;
	v->print_mi = NULL;
	v->new_viewer_mi = NULL;

	/* Root window submenu. */
	if(v->root_window_submenu != NULL)
	{
	    gtk_widget_destroy(v->root_window_submenu);
	    v->root_window_submenu = NULL;
	}

	/* Destroy image viewer. */
	ImgViewDelete(iv);
	v->iv = iv = NULL;


	/* Deallocate resources on the viewer structure. */

	/* Deallocate image info structure members. */
	clear_info(&v->info);

	/* Deallocate viewer structure itself. */
	g_free(v);
}

/*
 *	Destroys the grab window on the given core structure.
 */
static void grab_window_destroy(iv_core_struct *core_ptr)
{
	GtkWidget *w;

	if(core_ptr == NULL)
	    return;

	w = core_ptr->grab_window;
	core_ptr->grab_window = NULL;
	if(w != NULL)
	{
	    gtk_widget_destroy(w);
/* printf("Destroyed grab window 0x%.8x.\n", (guint)w); */
	    w = NULL;
	}
}


/*
 *	Update image info on given viewer structure with the info from
 *	the given imlib image.
 */
static void iv_update_image_info(
	iv_viewer_struct *v,
	const gchar *path,		/* Can be NULL. */
	struct stat *stat_buf,		/* Can be NULL. */
	gint width, gint height,
	gint bpl, gint bpp,
	gboolean has_alpha,
	const gchar *creator,		/* Can be NULL. */
        const gchar *author,            /* Can be NULL. */
        const gchar *comments,          /* Can be NULL. */
        gulong time_modified		/* In systime seconds. */
)
{
	iv_info_struct *info;


	if(v == NULL)
	    return;

	/* Get pointer to viewer's image info structure. */
	info = &v->info;

	/* Reset info structure. */
	clear_info(info);

	/* Update file name. */
	g_free(info->filename);
	info->filename = ((path != NULL) ? g_strdup(path) : NULL);

	/* Get file information if possible. */
	if(stat_buf != NULL)
	{
	    info->filesize = stat_buf->st_size;
	}


	/* Update the rest of the info. */
	info->width = width;
	info->height = height;
	info->has_alpha = has_alpha;

        /* Update creator only if changed. */
        if(creator != info->creator)
        {
            g_free(info->creator);
            info->creator = g_strdup((creator != NULL) ?
                creator : "Unknown"
            );
        }
        /* Update author only if changed. */
        if(author != info->author)
        {
            g_free(info->author);
            info->author = g_strdup((author != NULL) ?
                author : "Unknown"
            );
        }
	/* Update comments only if changed. */
	if(comments != info->comments)
	{
	    g_free(info->comments);
	    info->comments = g_strdup((comments != NULL) ?
		comments : NO_COMMENTS_STR
	    );
	}

	info->time_modified = time_modified;
}


/*
 *	Procedure to load an image to the given image viewer.
 */
static void do_load_image(
	iv_viewer_struct *v, const gchar *path, gboolean resize_on_load
)
{
	GtkWidget *w;
	gint i, status, width, height, bpl, bpp;
	guint8 *data_rgba;
	gchar *creator, *author, *comments;
	imgview_struct *iv;
	iv_imgio_cb_struct *cb_data;
	guint8 old_bg_color[4];
	struct stat stat_buf;


        if((v == NULL) || (path == NULL))
	    return;

	iv = v->iv;
	if(iv == NULL)
	    return;

        /* Get image viewer toplevel widget if it is a GtkWindow. */
	if(ImgViewToplevelIsWindow(iv))
	    w = ImgViewGetToplevelWidget(iv);
	else
	    w = NULL;

        /* Check if image file exists. */
        if(stat(path, &stat_buf))
	{
	    gchar *buf = g_strdup_printf(
		"No such file:\n\n    %s",
		path
	    );
            CDialogSetTransientFor(w);
            CDialogGetResponse(
                "Open Failed",
                buf,
"Please check if the specified image file path is\n\
correct and permissions allow reading of that file.",
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
	    g_free(buf);
	    return;
	}
#ifdef S_ISDIR
	/* Check if path is a directory. */
	if(S_ISDIR(stat_buf.st_mode))
	{
            gchar *buf = g_strdup_printf(
		"Cannot open directory:\n\n    %s",
		path
	    );
            CDialogSetTransientFor(w);
            CDialogGetResponse(
                "Open Failed",
                buf,
"The specified image file path reffers to a directory,\n\
directories cannot be opened as files.",
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
            g_free(buf);
            return;
	}
#endif  /* S_ISDIR */

	/* Record old background color. */
	memcpy(old_bg_color, v->bg_color, 4 * sizeof(guint8));

	/* Unload image on image viewer so that it is drawn as cleared.
	 * Note that the redraw event will be handled when the events
	 * are handled in the progress callback while loading the image
	 * just below.
	 */
	ImgViewClear(iv);


	/* Allocate load image callback data. */
	cb_data = (iv_imgio_cb_struct *)g_malloc0(sizeof(iv_imgio_cb_struct));
	if(cb_data != NULL)
	{
	    cb_data->ref_toplevel = w;
	    cb_data->viewer = v;
	    cb_data->path = g_strdup(path);
	}

	/* Load image. */
	data_rgba = NULL;
	creator = NULL;
	author = NULL;
	comments = NULL;
	status = ImgLoadFileRGBA(
	    path,
	    &width, &height,
	    &bpl, &bpp,
	    &data_rgba,
            v->bg_color,	/* 4 bytes in RGBA format, will be modified. */
	    &creator, &author, &comments,
	    0xff,		/* Default alpha value. */
	    cb_data,
	    iv_load_progress_cb
	);
	/* Note that data_rgba and comments now may be pointing to
	 * allocated memory, remember to deallocate them as needed.
	 */

	/* Deallocate load image callback data structure. */
	if(cb_data != NULL)
	{
	    g_free(cb_data->path);
	    g_free(cb_data);
	    cb_data = NULL;
	}
	/* Reset image viewer's progress bar. */
	ImgViewProgressUpdate(iv, 0.0, FALSE);


	/* Got image data? */
	if(data_rgba != NULL)
	{
	    /* Reset has changes mark. */
	    v->has_changes = FALSE;

	    /* Update image info on viewer structure with the info from
	     * the newly loaded image.
	     */
	    iv_update_image_info(
		v,
		path,
		&stat_buf,
		width, height, bpl, bpp,
		TRUE,
		creator, author, comments,
		MAX(stat_buf.st_mtime, stat_buf.st_ctime)
	    );

	    /* Update image viewer's background color. */
	    if(memcmp(old_bg_color, v->bg_color, 4 * sizeof(guint8)))
	    {
		GdkColor c[5];

                for(i = 0; i < 5; i++)
                {
                    c[i].red = (guint16)v->bg_color[0] << 8;
                    c[i].green = (guint16)v->bg_color[1] << 8;
                    c[i].blue = (guint16)v->bg_color[2] << 8;
                    c[i].pixel = 0;
                }
		ImgViewSetViewBG(iv, c);
	    }

	    /* Image viewer toplevel window available? */
	    if(w != NULL)
	    {
		/* Image viewer's toplevel widget is a GtkWindow, now
		 * we know we can resize it. Get the view drawing area
		 * and set its size to match the new image's size.
		 */
		GtkWidget *w2 = (GtkWidget *)ImgViewGetViewWidget(iv);
		if((w2 != NULL) && resize_on_load)
		{
		    gtk_widget_set_usize(
                        w2, width, height
		    );
		    gtk_widget_queue_resize(w2);
		    /* Notify toplevel to resize. */
		    gtk_widget_queue_resize(w);
		}

		/* Update viewer's toplevel window title. */
		update_view_title(v, path);
	    }

            /* Load new image to image viewer, any image currently loaded
	     * on the viewer will be unloaded first (there shouldn't
	     * be any since ImgViewClear() was called earlier).
             */
	    ImgViewLoad(
		iv,
		width, height, bpl,
		IMGVIEW_FORMAT_RGBA,
		(const guint8 *)data_rgba
	    );
        }
	else
	{
	    /* Failed to load image. */
	    gchar *buf = g_strdup_printf(
		"Cannot open image:\n\n    %s",
		path
	    );
	    const gchar *error_str = ImgLoadGetError();
            CDialogSetTransientFor(w);
            CDialogGetResponse(
		"Open Failed",
		buf,
		(error_str != NULL) ? error_str :
"Please check if the specified image file exists and\n\
is a valid format supported by Imlib.",
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
	    g_free(buf);
	}


	/* Deallocate the loaded image data and related resources. */
	g_free(data_rgba);
	data_rgba = NULL;

	g_free(creator);
	creator = NULL;

	g_free(author);
	author = NULL;

	g_free(comments);
	comments = NULL;


	/* Update menus. */
	update_view_menus(v);
}

/*
 *	Procedure to save an image to a different format.
 */
static void do_save_image(
	iv_viewer_struct *v, const gchar *path, const gchar *ext
)
{
        gint status;
        GtkWidget *w;
	iv_info_struct *info;
	imgview_struct *iv;
	iv_core_struct *core_ptr;
	
	gchar *orig_path = NULL;
	gchar *new_path = NULL;

	guint8 *src_data;
	gint width, height, bpl, bpp, format;

	struct stat stat_buf;
	iv_imgio_cb_struct *cb_data = NULL;


/* Macro to deallocate locally allocated resources. */
#define DO_FREE_LOCALS	\
{ \
 if(cb_data != NULL) \
 { \
  g_free(cb_data->path); \
  g_free(cb_data); \
  cb_data = NULL; \
 } \
\
 g_free(orig_path); \
 orig_path = NULL; \
\
 g_free(new_path); \
 new_path = NULL; \
}

        if((v == NULL) || (path == NULL))
	{
	    DO_FREE_LOCALS
            return;
	}
	if((*path) == '\0')
	{
            DO_FREE_LOCALS
            return;
	}

	/* Get pointers to image viewer and core structure. */
        iv = v->iv;
	core_ptr = (iv_core_struct *)v->core_ptr;
        if((iv == NULL) || (core_ptr == NULL))
	{
	    DO_FREE_LOCALS
            return;
	}
	info = &v->info;

	/* Check if image viewer does not have any image data loaded. */
	if(!ImgViewIsLoaded(iv))
	{
	    /* No image loaded, so nothing to save. */
	    DO_FREE_LOCALS
            return;
	}
	/* Get source image data and treat all returns as read-only. */
	src_data = ImgViewGetImageData(
	    iv, &width, &height, &bpl, &bpp, &format
	);


	/* If no extension given, then use generic string "*.*" as
	 * the extension.
	 */
	if(ext == NULL)
	    ext = "*.*";

	/* Begin generating new_path from given path and ext. */
	new_path = g_strdup(path);
	/* Special case for match any extension. */
	if(!strcmp(ext, "*.*") || !strcmp(ext, "*"))
	{
	    /* Leave new_path as is, but update ext to reflect
	     * actual extension.
	     */
	    ext = strrchr(path, '.');
	}
	else
	{
	    if(!extcmp(path, ext))
	    {
		gint	len1 = strlen(path),
			len2 = strlen(ext);

		new_path = (gchar *)g_realloc(
		    new_path,
		    (len1 + len2 + 1) * sizeof(gchar)
		);
		if(new_path != NULL)
		{
		    strcpy(new_path, path);
		    strcat(new_path, ext);
		}
	    }
	}
	/* Not able to generate new path? */
	if(new_path == NULL)
        {
            DO_FREE_LOCALS
            return;
        }

	/* At this point ext contains the string of the file type 
	 * extension, event if it was given as "*.*" it will be updated
	 * above to reflect the correct extension or ext can be NULL
	 * on failure.
	 */


        /* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;

	/* Check if file specified by new_path already exists. */
	if(!stat(new_path, &stat_buf))
	{
            /* Confirm overwrite. */
	    gchar *buf = g_strdup_printf(
		"Overwrite existing file:\n\n    %s",
		new_path
	    );
            CDialogSetTransientFor(w);
            status = CDialogGetResponse(
                "Confirm Overwrite",
                buf,
		NULL,
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
                CDIALOG_BTNFLAG_NO
            );
            CDialogSetTransientFor(NULL);
            g_free(buf);
	    switch(status)
	    {
	      case CDIALOG_RESPONSE_YES:
	      case CDIALOG_RESPONSE_YES_TO_ALL:
	      case CDIALOG_RESPONSE_OK:
		break;

	      default:
		DO_FREE_LOCALS
		return;
		break;
	    }
	}


        /* Get original file name from viewer structure, this may be 
	 * needed for some of the libraries. Note that orig_path may be
	 * NULL.
	 */
        orig_path = ((info->filename != NULL) ?
	    g_strdup(info->filename) : NULL
	);

	/* Allocate save callback data structure. */
        cb_data = (iv_imgio_cb_struct *)g_malloc0(sizeof(iv_imgio_cb_struct));
        if(cb_data != NULL)
        {
            cb_data->ref_toplevel = w;
            cb_data->viewer = v;
            cb_data->path = g_strdup(new_path);
        }


	/* Reset status to -1, it will be updated when (and if) an
	 * image saving routine is called.
	 */
	status = -1;

	/* Save file type by extension, check if we have a valid 
	 * extension.
	 */
	if(ext != NULL)
	{
	    gboolean got_response;

            /* GIF format? */
            if(!strcasecmp(ext, FTYPE_EXT_GIF))
            {
		/* Query user for GIF save options. */
                got_response = iv_query_gif_opts(
                    w,
                    new_path,
                    &core_ptr->gif_interlaced
                );
		/* Got positive user response and source image data format
		 * is RGBA?
		 */
		if(got_response && (format == IMGVIEW_FORMAT_RGBA))
		{
		    /* Save GIF using Imlib. */
		    status = ImgWriteFileRGBA(
			new_path,
			width, height, bpl, bpp,
			src_data,	/* Source RGBA data. */
			v->bg_color,	/* Background color. */
			NULL,		/* Comments. */
			1.0,		/* Quality, 0.0 to 1.0. */
			1,		/* Save as color (1 = color). */
			0xff,		/* Default alpha value. */
			core_ptr->imlib_fallback,
			cb_data, iv_save_progress_cb
		    );
                    /* Save failed? */
                    if((status != 0) && (status != -4))
                    {
                        CDialogSetTransientFor(w);
                        CDialogGetResponse(
                            "Save Failed",
                            "Imlib failed to save the image.",
                            NULL,
                            CDIALOG_ICON_ERROR,
                            CDIALOG_BTNFLAG_OK,
                            CDIALOG_BTNFLAG_OK
                        );
                        CDialogSetTransientFor(NULL);
                    }
		}
            }
	    /* JPEG format? */
	    else if(!strcasecmp(ext, FTYPE_EXT_JPG) ||
	            !strcasecmp(ext, FTYPE_EXT_JPEG)
            )
	    {
                /* Query user for JPEG save options. */
		got_response = iv_query_jpeg_opts(
		    w,
		    new_path,
		    &core_ptr->jpeg_quality,	/* 0.0 to 1.0. */
		    &core_ptr->jpeg_color
		);
                /* Got positive user response and source image data format
                 * is RGBA?
                 */
                if(got_response && (format == IMGVIEW_FORMAT_RGBA))
                {
                    /* Save JPEG using Imlib. */
                    status = ImgWriteFileRGBA(
                        new_path,
                        width, height, bpl, bpp,
                        src_data,       /* Source RGBA data. */
			v->bg_color,	/* Background color. */
                        NULL,           /* Comments. */
			core_ptr->jpeg_quality,	/* Quality, 0.0 to 1.0. */
                        (core_ptr->jpeg_color) ? 1 : 0,
                        0xff,           /* Default alpha value. */
                        core_ptr->imlib_fallback,
                        cb_data, iv_save_progress_cb
                    );
                    /* Save failed? */
                    if((status != 0) && (status != -4))
                    {
                        CDialogSetTransientFor(w);
                        CDialogGetResponse(
                            "Save Failed",
                            "Imlib failed to save the image.",
                            NULL,
                            CDIALOG_ICON_ERROR,
                            CDIALOG_BTNFLAG_OK,
                            CDIALOG_BTNFLAG_OK
                        );
                        CDialogSetTransientFor(NULL);
                    }
                }
	    }
            /* PNG format? */
            else if(!strcasecmp(ext, FTYPE_EXT_PNG))
            {
                /* Query user for PNG save options. */
                got_response = iv_query_png_opts(
                    w,
                    new_path,
                    &core_ptr->png_compression,	/* 0 to 9. */
                    &core_ptr->png_format,	/* 0 = B&W
						 * 1 = Greyscale
						 * 2 = Greyscale Alpha
						 * 3 = RGB
						 * 4 = RGBA
						 */
		    &core_ptr->png_interlaced
                );
                /* Got positive user response and source image data
		 * format is RGBA?
                 */
                if(got_response && (format == IMGVIEW_FORMAT_RGBA))
		{
		    status = ImgWritePNGFileRGBA(
			new_path,
			width, height, bpl, bpp,
                        src_data,	/* Source RGBA data. */
			v->bg_color,
			PROG_NAME " " PROG_VERSION,
			info->author, info->comments,
			core_ptr->png_compression,	/* 0 to 9. */
			core_ptr->png_interlaced,
			core_ptr->png_format,
			cb_data, iv_save_progress_cb
		    );
                    /* Save failed? */
                    if((status != 0) && (status != -4))
                    {
			gchar *buf = g_strdup_printf(
			    "Cannot save image:\n\n    %s",
			    new_path
			);
                        CDialogSetTransientFor(w);
                        CDialogGetResponse(
                            "Save Failed",
                            buf,
                            ImgWriteGetError(),
                            CDIALOG_ICON_ERROR,
                            CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
                            CDIALOG_BTNFLAG_OK
                        );
                        CDialogSetTransientFor(NULL);
			g_free(buf);
                    }
		}
	    }
            /* TGA format? */
            else if(!strcasecmp(ext, FTYPE_EXT_TGA))
            {
                /* Query user for TGA save options. */
                got_response = iv_query_tga_opts(
                    w,
                    new_path,
                    &core_ptr->tga_format	/* 0 = Greyscale
                                                 * 1 = RGB
                                                 * 2 = RGBA
                                                 */
                );
                /* Got positive user response and source image data
		 * format is RGBA?
                 */
                if(got_response && (format == IMGVIEW_FORMAT_RGBA))
                {
                    status = ImgWriteTGAFileRGBA(
                        new_path,
                        width, height, bpl, bpp,
                        src_data,       /* Source RGBA data. */
                        v->bg_color,
                        core_ptr->tga_format,
                        cb_data, iv_save_progress_cb
                    );
                    /* Save failed? */
                    if((status != 0) && (status != -4))
                    {
                        gchar *buf = g_strdup_printf(
                            "Cannot save image:\n\n    %s",
                            new_path
                        );
                        CDialogSetTransientFor(w);
                        CDialogGetResponse(
                            "Save Failed",
                            buf,
                            ImgWriteGetError(),
                            CDIALOG_ICON_ERROR,
                            CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
                            CDIALOG_BTNFLAG_OK
                        );
                        CDialogSetTransientFor(NULL);
                        g_free(buf);
                    }
                }
            }
	    else
	    {
		/* All else try to save using Imlib. */
                if(format == IMGVIEW_FORMAT_RGBA)
		{
                    /* Save image using Imlib. */
                    status = ImgWriteFileRGBA(
                        new_path,
                        width, height, bpl, bpp,
                        src_data,       /* Source RGBA data. */
			v->bg_color,	/* Background color. */
                        NULL,           /* Comments. */
                        1.0,            /* Quality, 0.0 to 1.0. */
                        1,              /* Save as color (1 = color). */
                        0xff,           /* Default alpha value. */
                        core_ptr->imlib_fallback,
                        cb_data, iv_save_progress_cb
                    );
		    /* Save failed? */
		    if((status != 0) && (status != -4))
		    {
			gchar *buf = g_strdup_printf(
			    "Cannot save image:\n\n    %s",
			    new_path
			);
			const gchar *error_str = ImgWriteGetError();
			CDialogSetTransientFor(w);
			CDialogGetResponse(
			    "Save Failed",
			    buf,
			    (error_str != NULL) ? error_str :
"Please check if the image format and any given save\n\
options are supported by Imlib.",
                            CDIALOG_ICON_ERROR,
                            CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
                            CDIALOG_BTNFLAG_OK
                        );
                        CDialogSetTransientFor(NULL);
			g_free(buf);
		    }
		}
	    }
	}	/* if(ext != NULL) */

	/* If save was successful, then status will be 0 at this point. */

	/* Save successful? */
	if(status == 0)
	{
	    gchar *tmp_author, *tmp_comments;


	    /* Reset has changes on successful save. */
	    v->has_changes = FALSE;

	    /* Get stats for new image. */
	    if(stat(new_path, &stat_buf))
		memset(&stat_buf, 0x00, sizeof(struct stat));

	    /* Update image info on viewer structure with the info from
	     * the newly saved image. Note that we copy the info's
	     * comments in order to preserve it by passing a copy of it
	     * when the rest of the image info is updated.
	     */
	    tmp_author = ((info->author != NULL) ?
                g_strdup(info->author) : NULL
            );
	    tmp_comments = ((info->comments != NULL) ?
		g_strdup(info->comments) : NULL
	    );
	    iv_update_image_info(
		v,
		new_path,
		&stat_buf,
		width, height, bpl, bpp,
		TRUE,
		PROG_NAME " " PROG_VERSION,
		tmp_author,
		tmp_comments,
		time(NULL)
	    );
	    g_free(tmp_author);
	    tmp_author = NULL;
	    g_free(tmp_comments);
	    tmp_comments = NULL;

	    /* Update viewer's toplevel window title. */
	    update_view_title(v, new_path);
	}

        /* Reset image viewer's progress bar. */
        ImgViewProgressUpdate(iv, 0.0, FALSE);

	/* Update menus. */
	update_view_menus(v);

	/* Deallocate locally allocated resources. */
	DO_FREE_LOCALS
#undef DO_FREE_LOCALS
}


/*
 *      Loads the image and displays it directly on the root window
 *      centered.
 */
static void do_load_image_to_root_centered(const gchar *path)
{
        gint status, width, height, bpl, bpp;
        guint8 *data_rgba;
        gchar *creator, *author, *comments;
        guint8 bg_color[4];


        /* Load image data. */
        data_rgba = NULL;
        creator = NULL;
        author = NULL;
        comments = NULL;
        status = ImgLoadFileRGBA(
            path,
            &width, &height,
            &bpl, &bpp,
            &data_rgba,
            bg_color,           /* 4 bytes in RGBA format, will be modified. */
            &creator, &author, &comments,
            0xff,               /* Default alpha value. */
            NULL, NULL          /* No progress callback. */
        );

        /* Got image data? */
        if(data_rgba != NULL)
        {
            GdkWindow *root_win = (GdkWindow *)GDK_ROOT_PARENT();
            GdkGC *gc = gdk_gc_new(root_win);

            if((root_win != NULL) && (gc != NULL))
                iv_put_image_to_root_centered(
                    root_win, gc,
                    (const guint8 *)data_rgba,
                    width, height, bpl, bpp,
                    IMGVIEW_FORMAT_RGBA
                );

            /* Should flush all GTK+ events. */
            while(gtk_events_pending() > 0)
                gtk_main_iteration();

            if(gc != NULL)
                gdk_gc_unref(gc);
        }

        /* Deallocate the loaded image data and related resources. */
        g_free(data_rgba);
        data_rgba = NULL;

        g_free(creator);
        creator = NULL;

        g_free(author);
        author = NULL;

        g_free(comments);
        comments = NULL;
}

/*
 *      Loads the image and displays it directly on the root window
 *      to fit.
 */
static void do_load_image_to_root_tofit(const gchar *path)
{
        gint status, width, height, bpl, bpp;
        guint8 *data_rgba;
        gchar *creator, *author, *comments;
        guint8 bg_color[4];


        /* Load image data. */
        data_rgba = NULL;
        creator = NULL;
        author = NULL;
        comments = NULL;
        status = ImgLoadFileRGBA(
            path,
            &width, &height,
            &bpl, &bpp,
            &data_rgba,
            bg_color,           /* 4 bytes in RGBA format, will be modified. */
            &creator, &author, &comments,
            0xff,               /* Default alpha value. */
            NULL, NULL          /* No progress callback. */
        );

        /* Got image data? */
        if(data_rgba != NULL)
        {
            GdkWindow *root_win = (GdkWindow *)GDK_ROOT_PARENT();
            GdkGC *gc = gdk_gc_new(root_win);

            if((root_win != NULL) && (gc != NULL))
                iv_put_image_to_root_tofit(
                    root_win, gc,
                    (const guint8 *)data_rgba,
                    width, height, bpl, bpp,
                    IMGVIEW_FORMAT_RGBA
                );

            /* Should flush all GTK+ events. */
            while(gtk_events_pending() > 0)
                gtk_main_iteration();

            if(gc != NULL)
                gdk_gc_unref(gc);
        }

        /* Deallocate the loaded image data and related resources. */
        g_free(data_rgba);
        data_rgba = NULL;

        g_free(creator);
        creator = NULL;

        g_free(author);
        author = NULL;

        g_free(comments);
        comments = NULL;
}

/*
 *	Loads the image and displays it directly on the root window
 *	to cover.
 */
static void do_load_image_to_root_tocover(const gchar *path)
{
        gint status, width, height, bpl, bpp;
        guint8 *data_rgba;
        gchar *creator, *author, *comments;
        guint8 bg_color[4];


	/* Load image data. */
	data_rgba = NULL;
	creator = NULL;
	author = NULL;
	comments = NULL;
	status = ImgLoadFileRGBA(
	    path,
	    &width, &height,
	    &bpl, &bpp,
	    &data_rgba,
	    bg_color,		/* 4 bytes in RGBA format, will be modified. */
	    &creator, &author, &comments,
	    0xff,		/* Default alpha value. */
	    NULL, NULL		/* No progress callback. */
        );

        /* Got image data? */
        if(data_rgba != NULL)
        {
	    GdkWindow *root_win = (GdkWindow *)GDK_ROOT_PARENT();
	    GdkGC *gc = gdk_gc_new(root_win);

	    if((root_win != NULL) && (gc != NULL))
		iv_put_image_to_root_tocover(
		    root_win, gc,
		    (const guint8 *)data_rgba,
		    width, height, bpl, bpp,
		    IMGVIEW_FORMAT_RGBA
		);

	    /* Should flush all GTK+ events. */
	    while(gtk_events_pending() > 0)
		gtk_main_iteration();

	    if(gc != NULL)
		gdk_gc_unref(gc);
	}

        /* Deallocate the loaded image data and related resources. */
        g_free(data_rgba);
        data_rgba = NULL;

        g_free(creator);
        creator = NULL;

        g_free(author);
        author = NULL;

        g_free(comments);
        comments = NULL;
}

/*
 *      Loads the image and displays it directly on the root window
 *      tiled.
 */
static void do_load_image_to_root_tiled(const gchar *path)
{
        gint status, width, height, bpl, bpp;
        guint8 *data_rgba;
        gchar *creator, *author, *comments;
        guint8 bg_color[4];


        /* Load image data. */
        data_rgba = NULL;
        creator = NULL;
        author = NULL;
        comments = NULL;
        status = ImgLoadFileRGBA(
            path,
            &width, &height,
            &bpl, &bpp,
            &data_rgba,
            bg_color,           /* 4 bytes in RGBA format, will be modified. */
            &creator, &author, &comments,
            0xff,               /* Default alpha value. */
            NULL, NULL          /* No progress callback. */
        );

        /* Got image data? */
        if(data_rgba != NULL)
        {
            GdkWindow *root_win = (GdkWindow *)GDK_ROOT_PARENT();
            GdkGC *gc = gdk_gc_new(root_win);

            if((root_win != NULL) && (gc != NULL))
                iv_put_image_to_root_tiled(
                    root_win, gc,
                    (const guint8 *)data_rgba,
                    width, height, bpl, bpp,
                    IMGVIEW_FORMAT_RGBA
                );

            /* Should flush all GTK+ events. */
            while(gtk_events_pending() > 0)
                gtk_main_iteration();

            if(gc != NULL)
                gdk_gc_unref(gc);
        }

        /* Deallocate the loaded image data and related resources. */
        g_free(data_rgba);
        data_rgba = NULL;

        g_free(creator);
        creator = NULL;

        g_free(author);
        author = NULL;

        g_free(comments);
        comments = NULL;
}


/*
 *	Image viewer image changed callback.
 *
 *	This is called whenever the image viewer has changed its image
 *	data from its own side invoked by the user.
 */
static void imgview_changed_cb(
	gpointer imgview, imgview_image_struct *image, gpointer data
)
{
	static gboolean reenterent = FALSE;
	gboolean size_changed = FALSE;
	GtkWidget *w, *w2;
	imgview_struct *iv = (imgview_struct *)imgview;
	iv_viewer_struct *v = (iv_viewer_struct *)data;
	iv_core_struct *core_ptr;
	if((iv == NULL) || (image == NULL) || (v == NULL))
	    return;

	core_ptr = (iv_core_struct *)v->core_ptr;
	if(core_ptr == NULL)
	    return;

	if(reenterent)
	    return;
	else
	    reenterent = TRUE;

	/* Update image information on viewer structure with the 
	 * information from the given image.
	 */
	if((v->info.width != image->width) ||
           (v->info.height != image->height)
	)
	    size_changed = TRUE;
	v->has_changes = TRUE;
	v->info.width = image->width;
	v->info.height = image->height;

        /* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;

	if(w != NULL)
	{
	    /* Get view widget from image viewer and resize it to match
	     * the new image.
	     */
	    w2 = (GtkWidget *)ImgViewGetViewWidget(iv);
	    if((w2 != NULL) && core_ptr->resize_on_load && size_changed)
	    {
		gtk_widget_set_usize(
		    w2, v->info.width, v->info.height
		);
		gtk_widget_queue_resize(w2);
		/* Notify toplevel to resize. */
		gtk_widget_queue_resize(w);
	    }

	    /* Update viewer's toplevel window title. */
	    update_view_title(v, NULL);
	}

	reenterent = FALSE;
}

/*
 *	Set view background color callback.
 */
static void set_bg_cb(GtkWidget *widget, gpointer data)
{
        static gboolean reenterent = FALSE;
	GtkWidget *w;
        csd_color_struct start_color, *color_rtn = NULL;
	imgview_struct *iv;
        iv_core_struct *core_ptr;
        iv_viewer_struct *v = (iv_viewer_struct *)data;
        if(v == NULL)
            return;

        if(v->processing)
            return;

        if(CSDIsQuery())
            return;

        iv = v->iv;
        core_ptr = (iv_core_struct *)v->core_ptr;
        if((iv == NULL) || (core_ptr == NULL))
            return;

        /* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;

        if(reenterent)
            return;
        else
            reenterent = TRUE;

	/* Set up initial color. */
	start_color.r = (gdouble)v->bg_color[0] / (gdouble)0xff;
        start_color.g = (gdouble)v->bg_color[1] / (gdouble)0xff;
        start_color.b = (gdouble)v->bg_color[2] / (gdouble)0xff;
        start_color.a = (gdouble)v->bg_color[3] / (gdouble)0xff;

        /* Prompt user to select new color. */
	CSDSetTransientFor(w);
        if(CSDGetResponse(
            "Set Background Color",
            "Set", "Cancel",
            &start_color, &color_rtn,
            NULL, NULL
        ))
        {
            if(color_rtn != NULL)
            {
                /* Got user response, set new background color. */
		gint i;
                GdkColor c[5];

		for(i = 0; i < 5; i++)
		{
		    c[i].red = (guint16)(color_rtn->r * 65535.0);
                    c[i].green = (guint16)(color_rtn->g * 65535.0);
                    c[i].blue = (guint16)(color_rtn->b * 65535.0);
		    c[i].pixel = 0;
		}
		ImgViewSetViewBG(iv, c);

		/* Record new background color on view structure. */
		v->bg_color[0] = (guint8)(color_rtn->r * 0xff);
                v->bg_color[1] = (guint8)(color_rtn->g * 0xff);
                v->bg_color[2] = (guint8)(color_rtn->b * 0xff);
                v->bg_color[3] = (guint8)(color_rtn->a * 0xff);
            }
        }
	CSDSetTransientFor(NULL);

	reenterent = FALSE;
}

/*
 *	Grab window callback.
 */
static void grab_start_cb(GtkWidget *widget, gpointer data)
{
	static gboolean reenterent = FALSE;
	gint status;
	gint width, height;
	GtkWidget *w, *w2;
	GtkWindow *grab_window;
        imgview_struct *iv;
	iv_info_struct *info;
        iv_core_struct *core_ptr;
        iv_viewer_struct *v = (iv_viewer_struct *)data;
        if(v == NULL)
            return;

        if(v->processing)
            return;

	info = &v->info;
        iv = v->iv;
        core_ptr = (iv_core_struct *)v->core_ptr;
        if((iv == NULL) || (core_ptr == NULL))
            return;

	/* Get grab window (may be NULL). */
	grab_window = (GtkWindow *)core_ptr->grab_window;

        /* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;

	/* Get image viewer's view widget. */
	w2 = (GtkWidget *)ImgViewGetViewWidget(iv);
	if(w2 == NULL)
	    return;


	if(reenterent)
	    return;
	else
	    reenterent = TRUE;


	/* Check for unsaved changes. */
	if(v->has_changes)
	{
            CDialogSetTransientFor(w);
            status = CDialogGetResponse(
"Save Changes?",
"The image data contains changes that have not been\n\
saved, do you want to save those changes?",
"There are changes made to the image data on this\n\
Image Viewer that have not been saved. If you want\n\
to save those changes then click on `Yes'. If you\n\
want to discard those changes then click on `No'.\n\
If you decide to not grab the image then click on\n\
`Cancel'.",
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO |
		CDIALOG_BTNFLAG_CANCEL | CDIALOG_BTNFLAG_HELP,
                CDIALOG_BTNFLAG_YES
            );
            CDialogSetTransientFor(NULL);
            switch(status)
            {
	      case CDIALOG_RESPONSE_CANCEL:
		reenterent = FALSE;
		return;
		break;

              case CDIALOG_RESPONSE_NO:
                /* Discard changes. */
                break;

              default:
                /* Call save image callback. */
                save_image_cb(v->save_as_mi, v);
                /* Check if save failed by checking if has_changes is
                 * still TRUE.
                 */
                if(v->has_changes)
                {
		    reenterent = FALSE;
                    return;
                }
                break;
            }
	}


	/* Get grab window options. */
	if(!iv_query_grab_opts(
	    w,
	    &core_ptr->grab_delay,	/* In milliseconds. */
	    &core_ptr->grab_hide_windows,
	    &core_ptr->grab_include_decorations
	))
	{
	    /* User canceled. */
	    reenterent = FALSE;
	    return;
	}


	/* Create grab window on core structure as needed. */
	if(grab_window == NULL)
	{
	    GtkWindow *gw = (GtkWindow *)gtk_window_new(
		GTK_WINDOW_POPUP
	    );
	    if(gw != NULL)
	    {
		gtk_widget_add_events(
		    GTK_WIDGET(gw),
		    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		    GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK
		);
		gtk_window_set_policy(gw, TRUE, TRUE, TRUE);
		gtk_widget_set_uposition(GTK_WIDGET(gw), 0, 0);
		gtk_widget_set_usize(GTK_WIDGET(gw), 1, 1);
                gtk_widget_realize(GTK_WIDGET(gw));
	    }
	    /* Record new grab window. */
	    core_ptr->grab_window = (GtkWidget *)grab_window = gw;
	}
	/* Not able to create grab window? */
	if(grab_window == NULL)
	{
            reenterent = FALSE;
            return;
	}


        /* Unmap image viewer for grab? */
        if(iv->map_state && core_ptr->grab_hide_windows)
            ImgViewUnmap(iv);

	/* Map grab window. */
	gtk_widget_show(GTK_WIDGET(grab_window));


	/* Need to flush GTK+ events, so new grab window creation and
	 * unmapping of other windows are realized.
	 */
	gdk_flush();
	while(gtk_events_pending() > 0)
	    gtk_main_iteration();

	/* Do grab. */
	status = iv_grab_start(
	    iv, &width, &height,
	    core_ptr->grab_delay,
	    core_ptr->grab_include_decorations,
	    core_ptr->resize_on_load,
	    GTK_WIDGET(grab_window),
	    core_ptr->grab_cur
	);

	/* Remap image viewer as needed. */
	if(!iv->map_state)
	    ImgViewMap(iv);

	/* Grab error, user aborted? */
	if(status == -4)
	{
	    /* User aborted, do nothing. */
	}
	/* Grab error, other error? */
	else if(status)
	{
	    /* Destroy the grab window, it's probably messed up. */
	    grab_window_destroy(core_ptr);
	    grab_window = NULL;

	    /* Show grab error message to user. */
	    CDialogSetTransientFor(w);
	    CDialogGetResponse(
		"Grab Window Failed",
		iv_grab_get_error(),
		NULL,
		CDIALOG_ICON_ERROR,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	}
	/* Grab successful. */
	else
	{
	    /* Unmap grab window, it is no longer needed. */
	    gtk_widget_hide(GTK_WIDGET(grab_window));


	    /* Increment untitled count on core structure since a new
	     * grabbed image was obtained.
	     */
	    core_ptr->untitled_count++;


	    /* Begin updating image information. */

	    /* Mark new grabbed image as not being saved yet. */
	    v->has_changes = TRUE;

	    /* Clear image info structure. */
	    clear_info(info);

	    info->width = width;
	    info->height = height;

	    info->has_alpha = FALSE;

	    /* Make a note that this is a grabbed image. */
	    g_free(info->comments);
	    info->comments = g_strdup("Came from a grabbed window");

	    /* Update viewer's toplevel window title. */
	    update_view_title(v, NULL);
	}

        /* Update menus. */
        update_view_menus(v);

	reenterent = FALSE;
}


/*
 *	Load progress callback, called by ImgIO*() functions when loading
 *	an image.
 */
gint iv_load_progress_cb(
        gpointer data,                  /* Client data. */
        gint min, gint max,             /* Progress. */
        gint width, gint height,
        gint bpl, gint bpp,
	guint8 *rgba_data
)
{
	static gboolean reenterent = FALSE;
	iv_viewer_struct *v;
	iv_imgio_cb_struct *cb_data = (iv_imgio_cb_struct *)data;
	if(cb_data == NULL)
	    return(FALSE);

	if(reenterent)
	    return(TRUE);
	else
	    reenterent = TRUE;

	/* Get viewer structure from callback data. */
	v = cb_data->viewer;
	if(v != NULL)
	{
	    /* Update image viewer's progress bar. */
	    if(max > 0)
		ImgViewProgressUpdate(
		    v->iv,
		    (gdouble)min / (gdouble)max,
		    TRUE
		);
	}

	reenterent = FALSE;
	return(TRUE);
}

/*
 *      Save progress callback, called by ImgIO*() functions when loading
 *      an image.
 */
static gint iv_save_progress_cb(
        gpointer data,                  /* Client data. */
        gint min, gint max,             /* Progress. */
        gint width, gint height,
        gint bpl, gint bpp,
        guint8 *rgba_data
)
{
        static gboolean reenterent = FALSE;
        iv_viewer_struct *v;
        iv_imgio_cb_struct *cb_data = (iv_imgio_cb_struct *)data;
        if(cb_data == NULL)
            return(FALSE);

        if(reenterent)
            return(TRUE);
        else
            reenterent = TRUE;

        /* Get viewer structure from callback data. */
        v = cb_data->viewer;
        if(v != NULL)
        {
            /* Update image viewer's progress bar. */
            if(max > 0)
                ImgViewProgressUpdate(
                    v->iv,
                    (gdouble)min / (gdouble)max,
                    TRUE
                );
        }

        reenterent = FALSE;
        return(TRUE);
}


/*
 *	Open image callback.
 */
static void open_image_cb(GtkWidget *widget, gpointer data)
{
	static gboolean reenterent = FALSE;
	gint status;
	gchar **path_rtn; gint path_total_rtns;
	fb_type_struct *type_rtn;
	gchar *strptr;
	fb_type_struct **ftype = NULL; gint total_ftypes = 0;
	GtkWidget *w;
	gchar tmp_path[PATH_MAX + NAME_MAX];
	imgview_struct *iv;
	iv_core_struct *core_ptr;
	iv_viewer_struct *v = (iv_viewer_struct *)data;
        if(v == NULL)
            return;

        if(v->processing)
            return;

	iv = v->iv;
	core_ptr = (iv_core_struct *)v->core_ptr;
        if((iv == NULL) || (core_ptr == NULL))
            return;

	if(FileBrowserIsQuery())
	    return;

        /* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;

	if(reenterent)
	    return;
	else
	    reenterent = TRUE;

	/* Check if there are unsaved changes. */
	if(v->has_changes)
	{
            CDialogSetTransientFor(w);
            status = CDialogGetResponse(
"Save Changes?",
"The image data contains changes that have not been\n\
saved, do you want to save those changes?",
"There are changes made to the image data on this\n\
Image Viewer that have not been saved. If you want\n\
to save those changes then click on `Yes'. If you\n\
want to discard those changes then click on `No'.\n\
If you decide to not open an image then click on\n\
`Cancel'.",
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO |
		CDIALOG_BTNFLAG_CANCEL | CDIALOG_BTNFLAG_HELP,
                CDIALOG_BTNFLAG_YES
            );
            CDialogSetTransientFor(NULL);
            switch(status)
            {
              case CDIALOG_RESPONSE_CANCEL:
		reenterent = FALSE;
		return;
                break;

              case CDIALOG_RESPONSE_NO:
		/* Discard changes. */
                break;

              default:
		/* Call save image callback. */
		save_image_cb(v->save_as_mi, v);
		/* Check if save failed by checking if has_changes is
		 * still TRUE.
		 */
		if(v->has_changes)
		{
		    reenterent = FALSE;
		    return;
		}
                break;
            }
	}



	/* Allocate file types list. For opening this is simple, since
	 * we can open virtually any image Imlib supports without having
	 * to specify its type.
	 */
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    "*.*", "Images"
	);

	/* Get new path from user. */
	FileBrowserSetTransientFor(w);
	if(FileBrowserGetResponse(
	    "Open Image", "Open", "Cancel",
	    current_directory,		/* Initial path. */
	    ftype, total_ftypes,
	    &path_rtn, &path_total_rtns,
	    &type_rtn
	))
	{
	    if(path_total_rtns > 0)
		strptr = path_rtn[0];
	    else
		strptr = NULL;
	    if(strptr != NULL)
	    {
		strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
		tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';


		ImgViewSetBusy(iv, TRUE);
		v->processing = TRUE;

		/* Load image. */
		do_load_image(v, tmp_path, core_ptr->resize_on_load);

		v->processing = FALSE;
		ImgViewSetBusy(iv, FALSE);


		/* Update current directory. */
		strncpy(current_directory, tmp_path, PATH_MAX);
		current_directory[PATH_MAX - 1] = '\0';

		strptr = strrchr(current_directory, '/');
		if(strptr != NULL)
		    (*strptr) = '\0';
                else
                    getcwd(current_directory, PATH_MAX);

		if((*current_directory) == '\0')
		    strcpy(current_directory, "/");
		current_directory[PATH_MAX - 1] = '\0';
	    }
	}
	FileBrowserSetTransientFor(NULL);

	/* Deallocate file types list. */
	FileBrowserDeleteTypeList(ftype, total_ftypes);

	reenterent = FALSE;
}

/*
 *	Save image callback.
 */
static void save_image_cb(GtkWidget *widget, gpointer data)
{
	static gboolean reenterent = FALSE;
        gchar **path_rtn; gint path_total_rtns;
	const gchar *imlib_extensions[] = IMLIB_EXTENSIONS;
	gint i;
        fb_type_struct *type_rtn;
        gchar *strptr;
        fb_type_struct **ftype = NULL; gint total_ftypes = 0;
        GtkWidget *w;
        gchar tmp_path[PATH_MAX + NAME_MAX];
        imgview_struct *iv;
	iv_viewer_struct *v = (iv_viewer_struct *)data;
        if(v == NULL)
            return;

        if(v->processing)
            return;

        iv = v->iv;
        if(iv == NULL)
            return;

        if(FileBrowserIsQuery())
            return;


        if(reenterent)
            return;
        else
            reenterent = TRUE;


        /* Allocate file types list, get list from Imlib supported
	 * formats. Note the list of extensions are in pairs of
	 * strings, the last pair being NULL and NULL.
	 */
        FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    "*.*", "Automatically detect format"
	);
	for(i = 0; imlib_extensions[i] != NULL; i += 2)
	    FileBrowserTypeListNew(
		&ftype, &total_ftypes,
		imlib_extensions[i], imlib_extensions[i + 1]
	    );

        /* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;

        /* Get new path from user. */
        FileBrowserSetTransientFor(w);
        if(FileBrowserGetResponse(
            "Save Image As", "Save", "Cancel",
            current_directory,		/* Initial path. */
            ftype, total_ftypes,
            &path_rtn, &path_total_rtns,
            &type_rtn
        ))
        {
            if(path_total_rtns > 0)
                strptr = path_rtn[0];
            else
                strptr = NULL;
            if(strptr != NULL)
            {
                strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
                tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';


                ImgViewSetBusy(iv, TRUE);
                v->processing = TRUE;

                /* Save image. */
                do_save_image(
		    v, tmp_path,
		    (type_rtn != NULL) ? type_rtn->ext : "*.*"
		);

                v->processing = FALSE;
                ImgViewSetBusy(iv, FALSE);


                /* Update current directory. */
                strncpy(current_directory, tmp_path, PATH_MAX);
                current_directory[PATH_MAX - 1] = '\0';

                strptr = strrchr(current_directory, '/');
                if(strptr != NULL)
                    (*strptr) = '\0';
                else
                    getcwd(current_directory, PATH_MAX);

                if((*current_directory) == '\0')
                    strcpy(current_directory, "/");
                current_directory[PATH_MAX - 1] = '\0';
            }
        }
        FileBrowserSetTransientFor(NULL);

        /* Deallocate file types list. */
        FileBrowserDeleteTypeList(ftype, total_ftypes);

	reenterent = FALSE;
}

/*
 *	Display image on root window callbacks.
 */
static void display_on_root_centered_cb(GtkWidget *widget, gpointer data)
{
        const guint8 *src_data;
        gint width, height, bpl, bpp, format;
        imgview_struct *iv;
        GdkWindow *root_win;
        GtkStyle *style;
        GtkWidget *w;
        iv_viewer_struct *v = (iv_viewer_struct *)data;
        if(v == NULL)
            return;

        if(v->processing)
            return;

        iv = v->iv;
        if(iv == NULL)
            return;

        /* Get style of image viewer toplevel. */
        w = ImgViewGetToplevelWidget(iv);
        if(w != NULL)
            style = gtk_widget_get_style(w);
        else
            style = NULL;
        if(style == NULL)
            return;

        /* Get root window. */
        root_win = (GdkWindow *)GDK_ROOT_PARENT();
        if(root_win == NULL)
            return;


        ImgViewSetBusy(iv, TRUE);
        v->processing = TRUE;

        src_data = (const guint8 *)ImgViewGetImageData(
            iv, &width, &height, &bpl, &bpp, &format
        );

        iv_put_image_to_root_centered(
            root_win, style->bg_gc[GTK_STATE_NORMAL],
            src_data, width, height, bpl, bpp, format
        );

        v->processing = FALSE;
        ImgViewSetBusy(iv, FALSE);
}

static void display_on_root_tofit_cb(GtkWidget *widget, gpointer data)
{
	const guint8 *src_data;
	gint width, height, bpl, bpp, format;
        imgview_struct *iv;
	GdkWindow *root_win;
	GtkStyle *style;
	GtkWidget *w;
        iv_viewer_struct *v = (iv_viewer_struct *)data;
        if(v == NULL)
            return;

	if(v->processing)
	    return;

        iv = v->iv;
        if(iv == NULL)
            return;

	/* Get style of image viewer toplevel. */
	w = ImgViewGetToplevelWidget(iv);
	if(w != NULL)
	    style = gtk_widget_get_style(w);
	else
	    style = NULL;
	if(style == NULL)
	    return;

	/* Get root window. */
        root_win = (GdkWindow *)GDK_ROOT_PARENT();
        if(root_win == NULL)
            return;


	ImgViewSetBusy(iv, TRUE);
	v->processing = TRUE;

	src_data = (const guint8 *)ImgViewGetImageData(
	    iv, &width, &height, &bpl, &bpp, &format
	);

	iv_put_image_to_root_tofit(
	    root_win, style->bg_gc[GTK_STATE_NORMAL],
	    src_data, width, height, bpl, bpp, format
	);

	v->processing = FALSE;
        ImgViewSetBusy(iv, FALSE);
}

static void display_on_root_tocover_cb(GtkWidget *widget, gpointer data)
{
        const guint8 *src_data;
        gint width, height, bpl, bpp, format;
        imgview_struct *iv;
        GdkWindow *root_win;
        GtkStyle *style;
        GtkWidget *w;
        iv_viewer_struct *v = (iv_viewer_struct *)data;
        if(v == NULL)
            return;

        if(v->processing)
            return;

        iv = v->iv;
        if(iv == NULL)
            return;

        /* Get style of image viewer toplevel. */
        w = ImgViewGetToplevelWidget(iv);
        if(w != NULL)
            style = gtk_widget_get_style(w);
        else
            style = NULL;
        if(style == NULL)
            return;

        /* Get root window. */
        root_win = (GdkWindow *)GDK_ROOT_PARENT();
        if(root_win == NULL)
            return;


        ImgViewSetBusy(iv, TRUE);
        v->processing = TRUE;

        src_data = (const guint8 *)ImgViewGetImageData(
            iv, &width, &height, &bpl, &bpp, &format
        );

        iv_put_image_to_root_tocover(
            root_win, style->bg_gc[GTK_STATE_NORMAL],
            src_data, width, height, bpl, bpp, format
        );

        v->processing = FALSE;
        ImgViewSetBusy(iv, FALSE);
}

static void display_on_root_tiled_cb(GtkWidget *widget, gpointer data)
{
        const guint8 *src_data;
        gint width, height, bpl, bpp, format;
        imgview_struct *iv;
        GdkWindow *root_win;
        GtkStyle *style;
        GtkWidget *w;
        iv_viewer_struct *v = (iv_viewer_struct *)data;
        if(v == NULL)
            return;

        if(v->processing)
            return;

        iv = v->iv;
        if(iv == NULL)
            return;

        /* Get style of image viewer toplevel. */
        w = ImgViewGetToplevelWidget(iv);
        if(w != NULL)
            style = gtk_widget_get_style(w);
        else
            style = NULL;
        if(style == NULL)
            return;

        /* Get root window. */
        root_win = (GdkWindow *)GDK_ROOT_PARENT();
        if(root_win == NULL)
            return;


        ImgViewSetBusy(iv, TRUE);
        v->processing = TRUE;

        src_data = (const guint8 *)ImgViewGetImageData(
            iv, &width, &height, &bpl, &bpp, &format
        );

        iv_put_image_to_root_tiled(
            root_win, style->bg_gc[GTK_STATE_NORMAL],
            src_data, width, height, bpl, bpp, format
        );

        v->processing = FALSE;
        ImgViewSetBusy(iv, FALSE);
}

/*
 *	Clear image callback.
 */
static void clear_image_cb(GtkWidget *widget, gpointer data)
{
        static gboolean reenterent = FALSE;
	gint status;
	GtkWidget *w;
	imgview_struct *iv;
        iv_viewer_struct *v = (iv_viewer_struct *)data;
        if(v == NULL)
            return;

        iv = v->iv;
        if(iv == NULL)
            return;

        /* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;

        if(reenterent)
            return;
        else
            reenterent = TRUE;

        /* Check if there are unsaved changes. */
        if(v->has_changes)
        {
            CDialogSetTransientFor(w);
            status = CDialogGetResponse(
"Save Changes?",
"The image data contains changes that have not been\n\
saved, do you want to save those changes?",
"There are changes made to the image data on this\n\
Image Viewer that have not been saved. If you want\n\
to save those changes then click on `Yes'. If you\n\
want to discard those changes then click on `No'.\n\
If you decide to not clear the image then click on\n\
`Cancel'.",
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO |
                CDIALOG_BTNFLAG_CANCEL | CDIALOG_BTNFLAG_HELP,
                CDIALOG_BTNFLAG_YES
            );
            CDialogSetTransientFor(NULL);
            switch(status)
            {
              case CDIALOG_RESPONSE_CANCEL:
                reenterent = FALSE;
                return;
                break;

              case CDIALOG_RESPONSE_NO:
                /* Discard changes. */
                break;

              default:
                /* Call save image callback. */
                save_image_cb(v->save_as_mi, v);
                /* Check if save failed by checking if has_changes is
                 * still TRUE.
                 */
                if(v->has_changes)
                {
                    reenterent = FALSE;
                    return;
                }
                break;
            }
        }


	/* Begin clearing image. */

	/* Unload image on image viewer. */
	ImgViewClear(iv);

	/* Begin resetting information values on viewer structure. */
	v->has_changes = FALSE;

	/* Reset image info structure. */
	clear_info(&v->info);

	/* Update title. */
	update_view_title(v, NULL);

        /* Update menus on viewer. */
        update_view_menus(v);

        reenterent = FALSE;
}

/*
 *	Image information callback.
 */
static void information_cb(GtkWidget *widget, gpointer data)
{
	static gboolean reenterent = FALSE;
	gchar file_size_str[256];
        gint width, height, bpl, bpp, format;
	const gchar	*filename, *ext_ptr, *name_ptr,
			*time_modified_str,
			*creator, *author, *comments;
	const gchar *alpha_channel_str = NULL;
	guint vis_id = 0x00000000;
	const gchar *dpy_name, *vis_name = NULL;
	gint depth = 0;
	GtkWidget *w, *w2;
	iv_info_struct *info;
        imgview_struct *iv;
        iv_viewer_struct *v = (iv_viewer_struct *)data;
        if(v == NULL)
            return;

	/* Get pointer to image info structure. */
	info = &v->info;

	/* Get pointer to image viewer. */
        iv = v->iv;
        if(iv == NULL)
            return;

        /* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;

	/* Get viewer widget. */
	w2 = (GtkWidget *)ImgViewGetViewWidget(iv);
	if((w2 != NULL) ? !GTK_WIDGET_NO_WINDOW(w2) : 0)
	{
	    GdkWindow *window = w2->window;
	    if(window != NULL)
	    {
		GdkVisual *vis = gdk_window_get_visual(window);

		/* Get Visual information. */
		if(vis != NULL)
		{
		    vis_id = (guint)vis;
		    switch((gint)vis->type)
		    {
		      case GDK_VISUAL_STATIC_GRAY:
			vis_name = "StaticGray";
			break;
		      case GDK_VISUAL_GRAYSCALE:
			vis_name = "GreyScale";
			break;
		      case GDK_VISUAL_STATIC_COLOR:
			vis_name = "StaticColor";
			break;
		      case GDK_VISUAL_PSEUDO_COLOR:
			vis_name = "PseudoColor";
			break;
		      case GDK_VISUAL_TRUE_COLOR:
			vis_name = "TrueColor";
			break;
		      case GDK_VISUAL_DIRECT_COLOR:
			vis_name = "DirectColor";
			break;
		    }
		    depth = vis->depth;
		}

	    }
	}

	/* Check if image viewer actually has an image loaded. */
	if(!ImgViewIsLoaded(iv))
	    return;

	/* Get image information. */
	if(ImgViewGetImageData(
	    iv,
	    &width, &height, &bpl, &bpp,
	    &format
	) == NULL)
	    return;

	/* Get image viewer's alpha channel info. */
	if((format == IMGVIEW_FORMAT_RGBA) ||
	   (format == IMGVIEW_FORMAT_GREYSCALEA32)
	)
	{
	    /* There is an allocated alpha channel, but check if
	     * it has defined data.
	     */
	    if(iv->alpha_channel_flags & IMGVIEW_ALPHA_DEFINED)
		alpha_channel_str = "Allocated and Defined";
	    else
		alpha_channel_str = "Allocated but Uniform";
	}
	else
	{
	    alpha_channel_str = "None";

	}


	if(reenterent)
	    return;
	else
	    reenterent = TRUE;


	/* Get other information about image. */

	/* Get pointer to complete file name. */
	filename = (const gchar *)info->filename;

	/* Get extension portion of file name if possible. */
	if(filename != NULL)
	    ext_ptr = (const gchar *)strrchr(filename, '.');
	else
	    ext_ptr = NULL;

	/* Get file's name without parent path if possible. */
        if(filename != NULL)
            name_ptr = (const gchar *)strrchr(filename, '/');
        else
            name_ptr = NULL;
	if(name_ptr == NULL)
	    name_ptr = filename;
	else
	    name_ptr++;

	/* Get last modified time string, note this string will be
	 * expected to have a '\n' character at the end (because
	 * ctime() tacks one on it.
	 */
	if(info->time_modified > 0)
	{
	    time_t t = info->time_modified;
	    time_modified_str = (const char *)ctime(&t);
	}
	else
	{
	    time_modified_str = "N/A\n";
	}

	/* Get creator, author, and comments. */
	creator = (const gchar *)info->creator;
	author = (const gchar *)info->author;
	comments = (const gchar *)info->comments;

	/* Set file size string. */
	if(info->filesize > 0)
	    sprintf(file_size_str, "%ld bytes", info->filesize);
	else
	    strcpy(file_size_str, "N/A");

	/* Get display name. */
	dpy_name = (const gchar *)gdk_get_display();

	if(1)
	{
	    /* Format message. */
	    gchar *buf = g_strdup_printf(
"\
Memory:\n\
Width: %i pixel%s\n\
Height: %i pixel%s\n\
Depth: %i byte%s per pixel\n\
Line: %i byte%s\n\
Size: %i bytes\n\
Alpha Channel: %s\n\
Has Changes: %s\n\
\n\
File: %s\n\
Format: %s\n\
Size: %s\n\
Last Modified: %s\
Creator: %s\n\
Author: %s\n\
Comments: %s\n\
\n\
Display: %s\n\
Visual Type: %s\n\
Depth: %i bits per pixel",
		/* Memory: */
		width, ((width != 1) ? "s" : ""),
		height, ((height != 1) ? "s" : ""),
		bpp, ((bpp != 1) ? "s" : ""),
		bpl, ((bpl != 1) ? "s" : ""),
		bpl * height,
		alpha_channel_str,
		(v->has_changes) ? "Yes" : "No",
		/* File: */
		(name_ptr != NULL) ? name_ptr : "N/A",
		(ext_ptr != NULL) ? ext_ptr : "N/A",
		file_size_str,
		time_modified_str,
                (creator != NULL) ? creator : "N/A",
                (author != NULL) ? author : "N/A",
		(comments != NULL) ? comments : NO_COMMENTS_STR,
		/* Display: */
		(dpy_name != NULL) ? dpy_name : ":0.0",
		(vis_name != NULL) ? vis_name : "N/A",
		depth
	    );
	    CDialogSetTransientFor(w);
	    CDialogGetResponse(
		(name_ptr != NULL) ? name_ptr : "Information",
		buf,
		NULL,
		CDIALOG_ICON_INFO,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    g_free(buf);
	}

	reenterent = FALSE;
}


/*
 *	Print callback.
 */
static void print_cb(GtkWidget *widget, gpointer data)
{
        static gboolean reenterent = FALSE;
	GtkWidget *w;
	print_values_struct *pv;
        imgview_struct *iv;
	iv_core_struct *core_ptr;
        iv_viewer_struct *v = (iv_viewer_struct *)data;
        if(v == NULL)
            return;

	iv = v->iv;
	core_ptr = (iv_core_struct *)v->core_ptr;
	if((iv == NULL) || (core_ptr == NULL))
	    return;

        /* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;

        /* Get print values. */
        pv = &core_ptr->print_values;


	if(reenterent)
	    return;
	else
	    reenterent = TRUE;


	/* Set up print values. */

	/* Get last print command and file name (these will not be 
	 * modified by the print routines).
	 */
	pv->print_command = core_ptr->print_command;
	pv->file_name = core_ptr->print_file;

	/* Check if an image is loaded on the image viewer, if it
	 * is then update print output geometry with its values.
	 */
	if(ImgViewIsLoaded(iv))
	{
	    const guint8 *data;
	    gint width, height, bpl, bpp, format;

	    data = ImgViewGetImageData(
		iv, &width, &height, &bpl, &bpp, &format
	    );
	    if(data != NULL)
	    {
		pv->output_width = width;
		pv->output_height = height;
	    }
	}

	/* Get user response for print values. */
	if(PrintWinGetResponse(
	    iv,		/* Reference image viewer. */
	    w,		/* Reference toplevel. */
	    pv		/* Print values, modified by call. */
	))
	{
	    ImgViewSetBusy(iv, TRUE);
	    v->processing = TRUE;

	    /* Call print procedure. */
	    iv_print(w, iv, pv);

	    v->processing = FALSE;
	    ImgViewSetBusy(iv, FALSE);


	    /* Update last print command and file name. */
	    if((pv->print_command != NULL) &&
               (pv->print_command != core_ptr->print_command)
	    )
	    {
		g_free(core_ptr->print_command);
		core_ptr->print_command = g_strdup(pv->print_command);
	    }
            if((pv->file_name != NULL) &&
               (pv->file_name != core_ptr->print_file)
            )
            {
                g_free(core_ptr->print_file);
                core_ptr->print_file = g_strdup(pv->file_name);
            }
	}

        reenterent = FALSE;
}


/*
 *	Create new viewer callback.
 */
static void new_viewer(GtkWidget *widget, gpointer data)
{
	gint n;
	iv_viewer_struct *new_v;
	iv_core_struct *core_ptr;
	imgview_struct *iv;
	GtkWidget *w;
        iv_viewer_struct *v = (iv_viewer_struct *)data;
        if(v == NULL)
            return;

        if(v->processing)
            return;

	iv = v->iv;
	core_ptr = (iv_core_struct *)v->core_ptr;
        if((iv == NULL) || (core_ptr == NULL))
            return;

        /* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;

	new_v = do_new_viewer(
	    core_ptr,
	    &core_ptr->viewer, &core_ptr->total_viewers, &n,
	    iv->toolbar_map_state,
	    iv->show_values,
	    iv->statusbar_map_state,
	    iv->show_image_on_wm_icon,
	    iv->quality,
	    (core_ptr->geometry_uset) ? &core_ptr->geometry : NULL,
	    TRUE		/* Always load title when user wants
				 * to create a new viewer.
				 */
	);
	if(new_v == NULL)
	{
            CDialogSetTransientFor(w);
            CDialogGetResponse(
"Operation Failed",
"Could not create a new viewer.",
"Please check memory and GUI resource\n\
availability.",
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
	}

	/* Update menus on given viewer. */
	update_view_menus(v);
}

/*
 *	Updates the toplevel window title of the given viewer.
 *
 *	If new_path is not NULL then it will be used as the new file
 *	name instead of v->filename.
 */
static void update_view_title(iv_viewer_struct *v, const gchar *new_path)
{
	static gboolean reenterent = FALSE;
	GtkWidget *w;
	const gchar *path;
	gint len;
	gchar *title;
	imgview_struct *iv;
	iv_core_struct *core_ptr;


	if(v == NULL)
	    return;

	iv = v->iv;
	core_ptr = (iv_core_struct *)v->core_ptr;
	if((iv == NULL) || (core_ptr == NULL))
	    return;

        /* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;
	if(w == NULL)
	    return;

	/* Get pointer to the new path if available, otherwise get
	 * pointer to the path specified on the viewer indicating
	 * the loaded file (or none if it is NULL).
	 */
	if(new_path != NULL)
	    path = new_path;
	else
	    path = (const gchar *)v->info.filename;


	if(reenterent)
	    return;
	else
	    reenterent = TRUE;


	/* Allocate memory for title string. */
	len = 256 + ((path != NULL) ? strlen(path) : 0);
	title = (gchar *)g_malloc((len + 1) * sizeof(gchar));
	if(title != NULL)
	{
	    const gchar *path_ptr;
	    gchar untitled_str[256];


	    /* If a path is available, it implies a file is loaded so
	     * we print its file name (without parent path compoents).
	     */
	    if(path != NULL)
	    {
		/* Path is specified, get only the file name without
		 * parent path compoents.
		 */
		path_ptr = strrchr(path, '/');
		if(path_ptr == NULL)
		    path_ptr = path;
		else
		    path_ptr++;
	    }
	    else
	    {
		/* No path available, implying no image is loaded so set
		 * path_ptr to point to a `untitled' string to use.
		 */
		sprintf(
		    untitled_str,
		    "Untitled #%i",
		    core_ptr->untitled_count
		);
		path_ptr = (const gchar *)untitled_str;
	    }

	    /* Format title string. */
	    sprintf(
		title,
		PROG_NAME ": %s %s",
		path_ptr,
		(v->has_changes) ? HAS_CHANGES_STR : ""
	    );
	    /* Set title to image viewer's toplevel window. */
	    gtk_window_set_title(GTK_WINDOW(w), title);

	    /* Deallocate title string, it is no longer needed. */
	    g_free(title);
	    title = NULL;
	}

	reenterent = FALSE;
}

/*
 *	Updates local menu item widgets on the given viewer.
 */
static void update_view_menus(iv_viewer_struct *v)
{
	static gboolean reenterent = FALSE;
	gboolean sensitive;
	GtkWidget *w;
	imgview_struct *iv;


	if(v == NULL)
	    return;

	iv = v->iv;
	if(iv == NULL)
	    return;

	if(reenterent)
	    return;
	else
	    reenterent = TRUE;


	/* Begin updating menu items. */
#define DO_SET_SENSITIVITY	\
{ \
 if(w != NULL) \
  gtk_widget_set_sensitive(w, sensitive); \
}

        w = v->save_as_mi;
        sensitive = ImgViewIsLoaded(iv);
        DO_SET_SENSITIVITY
        w = v->information_mi;
        DO_SET_SENSITIVITY
        w = v->root_window_mi;
        DO_SET_SENSITIVITY
        w = v->clear_mi;
        DO_SET_SENSITIVITY
	w = v->print_mi;
        DO_SET_SENSITIVITY

#undef DO_SET_SENSITIVITY

	reenterent = FALSE;
}


/*
 *	Deallocates and resets all members of the given info
 *	structure but does not deallocate the structure itself.
 */
static void clear_info(iv_info_struct *info)
{
	if(info == NULL)
	    return;

	/* Deallocate resources allocated in the info structure's
	 * members.
	 */
	g_free(info->filename);
	g_free(info->creator);
	g_free(info->author);
	g_free(info->comments);

	/* Reset all other members. */
	memset(info, 0x00, sizeof(iv_info_struct));
}


/*
 *	Checks if ext is an extension (postfix) of s (case insensitive).
 */
static gboolean extcmp(const gchar *s, const gchar *ext)
{
	gint s_len, ext_len;
	const gchar *s_ptr, *ext_ptr;


	if((s == NULL) || (ext == NULL))
	    return(FALSE);

	/* Neither strings can be empty. */
	if(((*s) == '\0') || ((*ext) == '\0'))
	    return(FALSE);

	/* Get lengths. */
	s_len = strlen(s);
	ext_len = strlen(ext);

	/* Set check pointers to ends. */
	s_ptr = &s[s_len];
	ext_ptr = &ext[ext_len];

	/* See backwards from ends, stop when one or both pointers
	 * reach the beginning of the string (whichever comes first).
	 */
	while((s_ptr >= s) && (ext_ptr >= ext))
	{
	    /* Characters do not match? */
	    if(toupper(*s_ptr--) != toupper(*ext_ptr--))
		break;
	}

	/* Matched entire extension string? */
	if(ext_ptr < ext)
	    return(TRUE);
	else
	    return(FALSE);
}


/*
 *	Parses the parameter and value from an argument, returning a
 *	statically allocated string or a string pointing within arg for
 *	both the parm_rtn and val_rtn.
 *
 *	The given arg will not be modified.
 */
static void get_args_value(
	char *arg, char **parm_rtn, char **val_rtn
)
{
	gchar *strptr;
	static gchar parm_str[256];


	if(parm_rtn != NULL)
	    (*parm_rtn) = "";
	if(val_rtn != NULL)
	    (*val_rtn) = "";

	if(arg == NULL)
	    return;

	/* If first char is not a '-' or '+' then skip as there is
	 * no parameter or value.
	 */
	if(((*arg) != '-') && ((*arg) != '+'))
	    return;

	/* Seek past initial '-' or '+' characters. */
	while(((*arg) == '-') || ((*arg) == '+'))
	    arg++;

	/* Get parameter and store it in our parm_str. */
	strncpy(parm_str, arg, 256);
	parm_str[256 - 1] = '\0';

	strptr = strchr(parm_str, '=');
	if(strptr != NULL)
	    (*strptr) = '\0';
	strptr = strchr(parm_str, '-');
        if(strptr != NULL)
            (*strptr) = '\0';

	/* Update parm return. */
	if(parm_rtn != NULL)
            (*parm_rtn) = parm_str;


	/* Get value. */
	strptr = strchr(arg, '=');
	if(strptr == NULL)
	    strptr = strchr(arg, '-');
	if(strptr == NULL)
	{
	    if(val_rtn != NULL)
                (*val_rtn) = "";
	}
	else
	{
            if(val_rtn != NULL)
                (*val_rtn) = strptr + 1;
	}
}

/*
 *	Returns TRUE if the given value is true, accepted values are
 *	"yes" "y" "on" "active" "enabled" or any non-zero value. Any
 *	other value will return FALSE.
 */
static gboolean stristrue(const gchar *s)
{
	gint c;


	if(s == NULL)
	    return(FALSE);

	c = (int)(*s);
	if((c == 'Y') || (c == 'y'))
	    return(TRUE);

	if(!strcasecmp(s, "on"))
	    return(TRUE);

	if(!strcasecmp(s, "active"))
            return(TRUE);

	if(!strcasecmp(s, "enabled"))
            return(TRUE);

	c = atoi(s);
	if(c != 0)
            return(TRUE);

	return(FALSE);
}

int main(int argc, char *argv[])
{
	gint rtn_val = 0;
	gboolean resize_on_load = TRUE;
	gboolean show_tooltips = TRUE;
	gboolean show_toolbar = TRUE;
	gboolean show_values = TRUE;
	gboolean show_statusbar = TRUE;
	gboolean show_image_on_wm_icon = FALSE;
	gint quality = 1;		/* 0 to 2. */
	gint imlib_fallback = 1;	/* 0 or 1. */
	const gchar *startup_file = NULL;
        const gchar *display_on_root_centered_path = NULL;
        const gchar *display_on_root_tofit_path = NULL;
        const gchar *display_on_root_tocover_path = NULL;
	const gchar *display_on_root_tiled_path = NULL;
	gboolean geometry_uset = FALSE;
	GdkRectangle geometry;

	gint i;
	gchar *strptr;
	const gchar *cstrptr;
	GdkCursor **cur;
	iv_viewer_struct *viewer;
	print_values_struct *pv;
	iv_core_struct *core_ptr;


        /* Parse arguments for help and version only. */
        for(i = 1; i < argc; i++)
        {
	    gchar *parm, *val;
            gchar *arg_ptr = (gchar *)argv[i];
            if(arg_ptr == NULL)
                continue;

            get_args_value(arg_ptr, &parm, &val);

            /* Help? */
            if(!strcasecmp(parm, "help") ||
               !strcasecmp(parm, "h") ||
               !strcasecmp(parm, "?")
            )
            {
                printf(PROG_USAGE);
                return(rtn_val);
            }
            /* Version? */
            else if(!strcasecmp(parm, "version"))
            {
                printf(
                    "%s Version %s\n%s",
                    PROG_NAME, PROG_VERSION,
                    PROG_COPYRIGHT
                );
                return(rtn_val);
            }
	}


        /* Reset time zone. */
        tzset();

        /* Get current directory. */
        if(getcwd(current_directory, PATH_MAX) == NULL)
           strcpy(current_directory, "/");
        current_directory[PATH_MAX - 1] = '\0';

	/* Set default geometry. */
	geometry.x = 0;
	geometry.y = 0;
	geometry.width = 100;
	geometry.height = geometry.width * 0.75;


	/* Initialize GTK+. */
        gtk_init(&argc, &argv);


	/* Parse arguments. */
	for(i = 1; i < argc; i++)
	{
	    gchar *parm, *val;
	    gchar *arg_ptr = (gchar *)argv[i];
	    if(arg_ptr == NULL)
		continue;

	    get_args_value(arg_ptr, &parm, &val);

	    /* General enable. */
            if(!strcasecmp(parm, "enable") ||
               !strcasecmp(parm, "with")
	    )
	    {
                if(!strcasecmp(val, "resize_on_load"))
                    resize_on_load = TRUE;
                else if(!strcasecmp(val, "show_image_on_wm_icon"))
                    show_image_on_wm_icon = TRUE;
                else if(!strcasecmp(val, "show_values"))
                    show_values = TRUE;
                else if(!strcasecmp(val, "show_toolbar"))
                    show_toolbar = TRUE;
		else if(!strcasecmp(val, "show_statusbar"))
		    show_statusbar = TRUE;
                else if(!strcasecmp(val, "show_tooltips"))
                    show_tooltips = TRUE;
            }
            /* General disable. */
            else if(!strcasecmp(parm, "disable") ||
                    !strcasecmp(parm, "without")
            )
            {
                if(!strcasecmp(val, "resize_on_load"))
                    resize_on_load = FALSE;
                else if(!strcasecmp(val, "show_image_on_wm_icon"))
                    show_image_on_wm_icon = FALSE;
                else if(!strcasecmp(val, "show_values"))
                    show_values = FALSE;
                else if(!strcasecmp(val, "show_toolbar"))
                    show_toolbar = FALSE;
		else if(!strcasecmp(val, "show_statusbar"))
		    show_statusbar = FALSE;
                else if(!strcasecmp(val, "show_tooltips"))
                    show_tooltips = FALSE;
            }
            /* Do not resize on load. */
            else if(!strcasecmp(parm, "r"))
            {
		resize_on_load = FALSE;
            }
	    /* Show/hide tooltips? */
	    else if(!strcasecmp(parm, "tooltips"))
            {
                if(stristrue(val))
		    show_tooltips = TRUE;
		else
		    show_tooltips = FALSE;
	    }
	    /* Hide tooltips? */
            else if(!strcasecmp(parm, "m"))
            {
		show_tooltips = FALSE;
            }
            /* Show/hide toolbar? */
            else if(!strcasecmp(parm, "toolbar"))
            {
                if(stristrue(val))
                    show_toolbar = TRUE;
                else
                    show_toolbar = FALSE;
            }
            else if(!strcasecmp(parm, "t"))
            {
		show_toolbar = FALSE;
	    }
            /* Show/hide values? */
            else if(!strcasecmp(parm, "show_values"))
            {
                if(stristrue(val))
                    show_values = TRUE;
                else
                    show_values = FALSE;
            }
            else if(!strcasecmp(parm, "v"))
            {
                show_values = FALSE;
            }
	    /* Show/hide progress/status bar? */
	    else if(!strcasecmp(parm, "show_statusbar"))
            {
                if(stristrue(val))
                    show_statusbar = TRUE;
                else
                    show_statusbar = FALSE;
            }
            else if(!strcasecmp(parm, "s"))
            {
                show_statusbar = FALSE;
            }
            /* Show image on wm icon? */
            else if(!strcasecmp(parm, "imageonwmicon"))
            {
		if(stristrue(val))
		    show_image_on_wm_icon = TRUE;
		else
		    show_image_on_wm_icon = FALSE;
            }
            else if(!strcasecmp(parm, "i"))
            {
		show_image_on_wm_icon = TRUE;
	    }
	    /* Quality. */
            else if(!strcasecmp(parm, "q"))
            {
		/* Value from 0 to 2 (2 being best/slowest). */
		quality = CLIP(atoi(val), 0, 2);
            }
            /* Imlib fallback. */
            else if(!strcasecmp(parm, "b"))
            {
		/* Value of 0 or 1. */
                imlib_fallback = CLIP(atoi(val), 0, 1);
            }
            /* Display image on root window? */
            else if(!strcasecmp(parm, "root"))
	    {
                if(*val == '\0')
                {
                    i++;
                    if(i < argc)
                    {
                        val = argv[i];
                    }
                    else
                    {
                        val = "";
                        fprintf(
                            stderr,
                            "%s: Requires argument.\n",
                            parm
                        );
                    }
                }
                if(*val != '\0')
                {
		    if(strstr(arg_ptr, "centered") ||
		       strstr(arg_ptr, "center")
		    )
			display_on_root_centered_path = val;
                    else if(strstr(arg_ptr, "tofit"))
                        display_on_root_tofit_path = val;
                    else if(strstr(arg_ptr, "tocover"))
                        display_on_root_tocover_path = val;
                    else if(strstr(arg_ptr, "tiled") ||
                            strstr(arg_ptr, "tile")
		    )
                        display_on_root_tiled_path = val;
                }
            }

	    /* Geometry. */
	    else if(!strcasecmp(parm, "geometry"))
            {
		if((*val) == '\0')
		{
		    i++;
		    if(i < argc)
		    {
		        val = argv[i];
		    }
		    else
		    {
			val = "";
			fprintf(
			    stderr,
			    "%s: Requires argument.\n",
			    parm
			);
		    }
		}
		if((*val) != '\0')
		{
		    gint rx, ry, mask;
		    guint rwidth, rheight;

		    mask = XParseGeometry(
			val, &rx, &ry, &rwidth, &rheight
		    );
		    if((mask & XValue) || (mask & XNegative))
			geometry.x = rx;
                    if((mask & YValue) || (mask & YNegative))
                        geometry.y = ry;
                    if(mask & WidthValue)
                        geometry.width = rwidth;
                    if(mask & HeightValue)
                        geometry.height = rheight;
		    if(mask)
		    {
			geometry_uset = TRUE;
			/* Setting geometry implies that we disable
			 * resize_on_load.
			 */
			resize_on_load = FALSE;
		    }
		}
	    }

	    /* File name (check if arg is not a toggle)? */
	    else if(((*arg_ptr) != '-') &&
                    ((*arg_ptr) != '+')
            )
            {
		startup_file = (const gchar *)arg_ptr;
            }
	    /* Unsupported option. */
	    else
	    {
		fprintf(stderr, "%s: Unsupported option.\n", parm);
		rtn_val = 1;
		return(rtn_val);
	    }
	}


	/* Allocate our core structure. */
	core_ptr = (iv_core_struct *)g_malloc0(sizeof(iv_core_struct));
	if(core_ptr == NULL)
	{
	    rtn_val = 3;
	    return(rtn_val);
	}
	/* Begin resetting values on the core structure. */
	core_ptr->viewer = NULL;
	core_ptr->total_viewers = 0;

	/* Reset untitled image count. */
	core_ptr->untitled_count = 0;

        /* Create cursors. */
	core_ptr->grab_cur = gdk_cursor_new(GDK_CROSSHAIR);

	/* Reset grab window, it will be created when the grab
	 * callback is called.
	 */
	core_ptr->grab_window = NULL;

	/* Options. */
	core_ptr->show_tooltips = show_tooltips;
	/* These values are record at startup, but are NOT updated
	 * during run time so they are not accurate at run time.
	 */
	core_ptr->show_toolbar = show_toolbar;
	core_ptr->show_values = show_values;
	core_ptr->show_statusbar = show_statusbar;
	core_ptr->show_image_on_wm_icon = show_image_on_wm_icon;
	core_ptr->quality = quality;

	/* Set geometry if defined. */
        core_ptr->geometry_uset = geometry_uset;
	if(core_ptr->geometry_uset)
	    memcpy(&core_ptr->geometry, &geometry, sizeof(GdkRectangle));

        /* Resize image viewer whenever a new image is loaded. */
	core_ptr->resize_on_load = resize_on_load;
	core_ptr->imlib_fallback = imlib_fallback;

	/* File format last save options. */
	core_ptr->gif_interlaced = GIF_DEF_INTERLACED;
	core_ptr->jpeg_quality = JPEG_DEF_QUALITY;
	core_ptr->jpeg_color = JPEG_DEF_IS_COLOR;
	core_ptr->png_compression = PNG_DEF_COMPRESSION;
	core_ptr->png_format = PNG_DEF_FORMAT;
	core_ptr->png_interlaced = PNG_DEF_INTERLACED;
	core_ptr->tga_format = TGA_DEF_FORMAT;

	/* Grab window options. */
	core_ptr->grab_delay = GRAB_DEF_DELAY;
	core_ptr->grab_include_decorations = GRAB_DEF_INCLUDE_DECORATIONS;
        core_ptr->grab_hide_windows = GRAB_DEF_HIDE_WINDOWS;

	/* Print options. */
	pv = &core_ptr->print_values;
        pv->flags = (PRINT_VALUE_OUTPUT_TYPE | PRINT_VALUE_PRINT_COMMAND |
                PRINT_VALUE_FILE_NAME | PRINT_VALUE_BRIGHTNESS |
                PRINT_VALUE_CONTRAST |
                PRINT_VALUE_VISUAL | PRINT_VALUE_ORIENTATION |
                PRINT_VALUE_DPI_X | PRINT_VALUE_DPI_Y |
                PRINT_VALUE_PAPER_WIDTH | PRINT_VALUE_PAPER_HEIGHT |
                PRINT_VALUE_OUTPUT_X | PRINT_VALUE_OUTPUT_Y |
                PRINT_VALUE_OUTPUT_WIDTH | PRINT_VALUE_OUTPUT_HEIGHT
        );
        pv->output_type = PRINT_OUTPUT_TO_PRINTER;
        pv->print_command = NULL;
        pv->file_name = NULL;
        pv->brightness = 0.0;
        pv->contrast = 0.0;
        pv->visual = PRINT_VISUAL_COLOR;
        pv->orientation = PRINT_ORIENTATION_PORTRAIT;
        pv->dpi_x = 72;
        pv->dpi_y = 72;
        pv->paper_width = 850;	/* Letter sized paper. */
        pv->paper_height = 1100;
        pv->output_x = 0;
        pv->output_y = 0;
        pv->output_width = 320;
        pv->output_height = 240;
	/* Default print command. */
	core_ptr->print_command = g_strdup(PRINT_DEF_COMMAND);
	/* Default print file name. */
	cstrptr = (const gchar *)getenv("HOME");
	if(cstrptr == NULL)
	    cstrptr = "";
	core_ptr->print_file = (gchar *)g_malloc(
	    strlen(cstrptr) + strlen(PRINT_DEF_FILE) + 10
	);
	if(core_ptr->print_file != NULL)
	{
	    strcpy(core_ptr->print_file, cstrptr);
	    strcat(core_ptr->print_file, "/");
	    strcat(core_ptr->print_file, PRINT_DEF_FILE);
	}

#define DO_DESTROY_CURSOR	\
{ \
 if((*cur) != NULL) \
 { \
  GdkCursor *tc = (*cur); \
  (*cur) = NULL; \
  gdk_cursor_destroy(tc); \
 } \
}

/* Macro to deallocate the core structure and all its resources, uses
 * variable core_ptr and global iv_core.
 */
#define DO_SHUTDOWN	\
{ \
 if(core_ptr != NULL) \
 { \
  /* Begin deallocating core structure. */ \
  /* Deallocate all viewer structures. */ \
  for(i = 0; i < core_ptr->total_viewers; i++) \
  { \
   viewer_delete(core_ptr->viewer[i]); \
   core_ptr->viewer[i] = NULL; \
  } \
  g_free(core_ptr->viewer); \
  core_ptr->viewer = NULL; \
  core_ptr->total_viewers = 0; \
  viewer = NULL; \
\
  /* Destroy the GtkWindow used in grab window operations. */ \
  grab_window_destroy(core_ptr); \
\
  /* Deallocate cursors. */ \
  cur = &core_ptr->grab_cur; \
  DO_DESTROY_CURSOR \
\
  /* Deallocate other core structure resources. */ \
  g_free(core_ptr->print_command); \
  core_ptr->print_command = NULL; \
  g_free(core_ptr->print_file); \
  core_ptr->print_file = NULL; \
\
  /* Deallocate core structure itself. */ \
  g_free(core_ptr); \
  core_ptr = NULL; \
 } \
\
 /* Shutdown our GUI utilities. */ \
 ProgressDialogShutdown(); \
 FileBrowserShutdown(); \
 CSDShutdown(); \
 FPromptShutdown(); \
 CDialogShutdown(); \
\
 /* GTK+ does not need to be shut down. */ \
}


	/* Initialize GDK RGB buffers. */
	gdk_rgb_init();

#ifdef HAVE_IMLIB
	/* Initialize Imlib. */
	imlib_handle = Imlib_init(GDK_DISPLAY());
#endif	/*HAVE_IMLIB */

	/* Update global tooltips state. */
	GUISetGlobalTipsState(core_ptr->show_tooltips);

	/* Initialize our GUI utilities. */
	CDialogInit();
	FPromptInit();
	CSDInit();
	FileBrowserInit();
	ProgressDialogInit();


	/* Check special startup cases first. */
	/* Display image directly on root window and exit? */
        if(display_on_root_centered_path != NULL)
        {
            do_load_image_to_root_centered(
                display_on_root_centered_path
            );
            DO_SHUTDOWN
            rtn_val = 0;
            return(rtn_val);
        }
        else if(display_on_root_tofit_path != NULL)
        {
            do_load_image_to_root_tofit(
                display_on_root_tofit_path
            );
            DO_SHUTDOWN
            rtn_val = 0;
            return(rtn_val);
        }
	else if(display_on_root_tocover_path != NULL)
        {
            do_load_image_to_root_tocover(
                display_on_root_tocover_path
            );
            DO_SHUTDOWN
            rtn_val = 0;
            return(rtn_val);
        }
	else if(display_on_root_tiled_path != NULL)
	{
	    do_load_image_to_root_tiled(
		display_on_root_tiled_path
	    );
	    DO_SHUTDOWN
	    rtn_val = 0;
	    return(rtn_val);
	}


	/* Create a new image viewer. */
	viewer = do_new_viewer(
	    core_ptr,
	    &core_ptr->viewer, &core_ptr->total_viewers, &i,
	    show_toolbar,
	    show_values,
	    show_statusbar,
	    show_image_on_wm_icon,
	    quality,
	    (geometry_uset) ? &geometry : NULL,
	    (startup_file == NULL) ? TRUE : FALSE
	);
	if(viewer != NULL)
	{
	    /* Load startup image if path is not NULL. */
	    if(startup_file != NULL)
	    {
                /* Update current directory. */
                strncpy(current_directory, startup_file, PATH_MAX);
		current_directory[PATH_MAX - 1] = '\0';
		strptr = strrchr(current_directory, '/');
		if(strptr != NULL)
		    (*strptr) = '\0';
		else
		    getcwd(current_directory, PATH_MAX);
		current_directory[PATH_MAX - 1] = '\0';

		/* Load image file. */
		do_load_image(viewer, startup_file, core_ptr->resize_on_load);
	    }
	}
	else
	{
	    DO_SHUTDOWN
	    rtn_val = 3;
	    return(rtn_val);
	}


	/* Enter a GTK+ management loop. */
	gtk_main();


        /* Deallocate core structure and all related resources. */
        DO_SHUTDOWN
#undef DO_SHUTDOWN
#undef DO_DESTROY_CURSOR

#ifdef HAVE_IMLIB
        /* Imlib does not need to be shut down. */
        imlib_handle = NULL;
#endif  /* HAVE_IMLIB */


	return(rtn_val);
}
