#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "guiutils.h"
#include "saveopts.h"

#include "images/icon_save_20x20.xpm"
#include "images/icon_cancel_20x20.xpm"
#include "saveopts.xpm"


static gint close_cb(GtkWidget *widget, GdkEvent *event, gpointer data);
static void destroy_cb(GtkObject *object, gpointer data);
static void save_btn_cb(GtkWidget *widget, gpointer data);
static void cancel_btn_cb(GtkWidget *widget, gpointer data);
static void scale_changed_cb(GtkAdjustment *adjustment, gpointer data);
static void entry_changed_adj_cb(GtkEntry *entry, gpointer data);

static void build_std_dialog(
        gint width, gint height, const gchar *title,
        GtkWidget *ref_toplevel,
        GtkWidget **toplevel_rtn, GtkWidget **client_vbox_rtn,
        GtkAccelGroup **accelgrp_rtn
);


gint iv_query_gif_opts(
        GtkWidget *ref_toplevel,
        const gchar *path,
        gbool *interlaced
);
gint iv_query_jpeg_opts(
	GtkWidget *ref_toplevel,
	const gchar *path,
	gfloat *quality,	/* Quality value from 0.0 to 1.0. */
	gbool *save_as_color
);
gint iv_query_png_opts(
        GtkWidget *ref_toplevel,
        const gchar *path,
        gint *compression,      /* Value from 0 to 9. */
        gint *format,           /* 0 = B&W  1 = Greyscale
                                 * 2 = Greyscale Alpha
                                 * 3 = RGB  4 = RGBA
                                 */
	gbool *interlaced
);
gint iv_query_tga_opts(
        GtkWidget *ref_toplevel,
        const gchar *path,
        gint *format		/* 0 Greyscale
				 * 1 = RGB
				 * 2 = RGBA
				 */
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))

/* Tooltip messages. */
#define SAVE_OPTS_TTM_JPEG_QUALITY	"Percentage of data retained, \
higher percentages improve the quality and reduce the amount of data \
discarded but increases the file size"

#define SAVE_OPTS_TTM_PNG_COMPRESSION	"Higher values increase the level \
of compression and time it takes to save, optimal compression is 6 and \
maximum is 9. PNG compression never discards data"
#define SAVE_OPTS_TTM_PNG_ALPHA		"Check to save the alpha \
channel, the alpha channel is used to store PNG transparency and \
blending information"
#define SAVE_OPTS_TTM_PNG_INTERLACED	"Check to save the image data \
using Adam 7 interlacing (no data will be discarded)"


#define BTN_WIDTH	(100 + (2 * 3))
#define BTN_HEIGHT	(30 + (2 * 3))


static gbool got_response;


/*
 *	Closes the given toplevel window given by data and breaks
 *	the highest gtk_main level.
 */
static gint close_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	cancel_btn_cb(widget, data);
	return(TRUE);
}

/*
 *	Destroy callback.
 */
static void destroy_cb(GtkObject *object, gpointer data)
{
	return;
}

/*
 *	Save button callback.
 */
static void save_btn_cb(GtkWidget *widget, gpointer data)
{
        GtkWidget *toplevel = (GtkWidget *)data;
        if(toplevel == NULL)
            return;

	got_response = TRUE;

        gtk_main_quit();
}

/*
 *	Cancel button callback.
 */
static void cancel_btn_cb(GtkWidget *widget, gpointer data)
{
        GtkWidget *toplevel = (GtkWidget *)data;
        if(toplevel == NULL)
            return;

        got_response = FALSE;

        gtk_main_quit();
}

/*
 *	Scale value changed.
 */
static void scale_changed_cb(GtkAdjustment *adjustment, gpointer data)
{
	static gbool reenterant = FALSE;
	GtkAdjustment *adj = adjustment;
	GtkWidget *w = (GtkWidget *)data;
	if((adj == NULL) || (w == NULL))
	    return;

	if(reenterant)
	    return;
	else
	    reenterant = TRUE;

	if(GTK_IS_ENTRY(w))
	{
	    gchar num_str[80];
	    GtkEntry *entry = GTK_ENTRY(w);
	    sprintf(num_str, "%.0f", adj->value);
	    gtk_entry_set_text(entry, num_str);
	}

	reenterant = FALSE;
}

/*
 *	Entry value changed with adjustment as data callback.
 */
void entry_changed_adj_cb(GtkEntry *entry, gpointer data)
{
        static gbool reenterant = FALSE;
	const gchar *cstrptr;
        GtkAdjustment *adj = (GtkAdjustment *)data;
        if((entry == NULL) || (adj == NULL))
            return;

        if(reenterant)
            return;
        else
            reenterant = TRUE;

	cstrptr = (const gchar *)gtk_entry_get_text(entry);
	if(cstrptr != NULL)
	{
	    adj->value = atof(cstrptr);
	    gtk_signal_emit_by_name(GTK_OBJECT(adj), "changed");
	}

	reenterant = FALSE;
}

/*
 *	Builds a standard toplevel dialog and returns the pointers to
 *	the created widgets.
 */
static void build_std_dialog(
	gint width, gint height, const gchar *title,
	GtkWidget *ref_toplevel,
	GtkWidget **toplevel_rtn, GtkWidget **client_vbox_rtn,
	GtkAccelGroup **accelgrp_rtn
)
{
	gint	bw = BTN_WIDTH,
		bh = BTN_HEIGHT,
		border_major = 5;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
        GtkWidget *w, *parent, *parent2;
        GtkWidget *toplevel, *main_vbox;


        /* Create keyboard accelerator group. */
        accelgrp = gtk_accel_group_new();
	if(accelgrp_rtn != NULL)
	    (*accelgrp_rtn) = accelgrp;

        /* Create toplevel. */
        toplevel = parent = w = gtk_window_new(GTK_WINDOW_DIALOG);
        gtk_widget_set_usize(w, width, height);
        gtk_widget_realize(w);
        gtk_window_set_title(GTK_WINDOW(w), title);
        window = w->window;
        if(window != NULL)
        {
            gdk_window_set_decorations(
                window,
                GDK_DECOR_TITLE | GDK_DECOR_MENU | GDK_DECOR_MINIMIZE
            );
            gdk_window_set_functions(
                window,
                GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
            );
            GUISetWMIcon(window, (u_int8_t **)saveopts_xpm);
        }
        gtk_signal_connect(
            GTK_OBJECT(w), "delete_event",
            GTK_SIGNAL_FUNC(close_cb),
            (gpointer)toplevel
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "destroy",
            GTK_SIGNAL_FUNC(destroy_cb),
            (gpointer)toplevel
        );
        gtk_container_border_width(GTK_CONTAINER(w), 0);
        gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
        if((ref_toplevel != NULL) ? GTK_IS_WINDOW(GTK_OBJECT(ref_toplevel)) : 0)
        {
            gtk_window_set_modal(GTK_WINDOW(toplevel), TRUE);
            gtk_window_set_transient_for(
                GTK_WINDOW(toplevel), GTK_WINDOW(ref_toplevel)
            );
        }
	if(toplevel_rtn != NULL)
	    (*toplevel_rtn) = toplevel;

        main_vbox = w = gtk_vbox_new(FALSE, 0);
        gtk_container_add(GTK_CONTAINER(parent), w);
        gtk_widget_show(w);
        parent = w;

        w = gtk_vbox_new(FALSE, border_major);
        gtk_container_border_width(GTK_CONTAINER(w), border_major);
        gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
        gtk_widget_show(w);
        parent2 = w;
	if(client_vbox_rtn != NULL)
	    (*client_vbox_rtn) = w;

	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, 0);
        gtk_widget_show(w);

        /* Buttons. */
        w = gtk_hbox_new(TRUE, 0);
        gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, border_major);
        gtk_widget_show(w);
        parent2 = w;

        /* Save button. */
        w = GUIButtonPixmapLabelH(
            (u_int8_t **)icon_save_20x20_xpm, "Save", NULL
        );
        gtk_widget_set_usize(w, bw, bh);
        GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT | GTK_HAS_DEFAULT);
        gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(save_btn_cb),
            (gpointer)toplevel
        );
        gtk_accel_group_add(
            accelgrp, GDK_Return, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_3270_Enter, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_KP_Enter, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_accel_group_add(
            accelgrp, GDK_ISO_Enter, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_widget_show(w);

        /* Cancel button. */
        w = GUIButtonPixmapLabelH(
            (u_int8_t **)icon_cancel_20x20_xpm, "Cancel", NULL
        );
        gtk_widget_set_usize(w, bw, bh);
        GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
        gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(cancel_btn_cb),
            (gpointer)toplevel
        );
        gtk_accel_group_add(
            accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
            GTK_OBJECT(w), "clicked"
        );
        gtk_widget_show(w);
}


/*
 *	GIF format options querying.
 *
 *      Blocks execution untill response is recieved.
 *      Returns TRUE if values were fetched or FALSE if user canceled.
 */
gint iv_query_gif_opts(
        GtkWidget *ref_toplevel,
        const gchar *path,
        gbool *interlaced
)
{
        gint    width = 320,
                height = -1,
                border_minor = 2;
        GtkAccelGroup *accelgrp = NULL;
        GtkWidget *toplevel = NULL, *client_vbox = NULL;
        GtkWidget *interlaced_toggle = NULL;


        /* Reset global got response flag. */
        got_response = FALSE;

        /* Create dialog base widgets. */
        build_std_dialog(
            width, height,
            "GIF Format Options",
            ref_toplevel,
            &toplevel, &client_vbox, &accelgrp
        );

        /* Create GIF format option widgets. */
        if(client_vbox != NULL)
        {
            GtkWidget *w, *parent = client_vbox, *parent2;

            /* Interlaced. */
            w = gtk_hbox_new(FALSE, border_minor);
            gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            parent2 = w;
            /* Check button. */
            interlaced_toggle = w = gtk_check_button_new_with_label("Interlaced");
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            if((interlaced != NULL) ? *interlaced : 0)
                GTK_TOGGLE_BUTTON(w)->active = TRUE;
            gtk_widget_show(w);

        }


        /* Map toplevel. */
        gtk_widget_show(toplevel);


        /* Enter blocking gtk_main loop. */
        gtk_main();


        /* Get values from widgets if user clicked on save. */
        if(got_response)
        {
            if((interlaced != NULL) && (interlaced_toggle != NULL))
                (*interlaced) = GTK_TOGGLE_BUTTON(interlaced_toggle)->active;
        }


        /* Destroy widgets and related resources. */
        gtk_window_set_modal(GTK_WINDOW(toplevel), FALSE);
        gtk_window_set_transient_for(GTK_WINDOW(toplevel), NULL);
        gtk_widget_destroy(toplevel);
        gtk_accel_group_unref(accelgrp);


        /* Return user response. */
        return(got_response);
}





/*
 *	JPEG format options querying.
 *
 *	Blocks execution untill response is recieved.
 *	Returns TRUE if values were fetched or FALSE if user canceled.
 */
gint iv_query_jpeg_opts(
        GtkWidget *ref_toplevel,
	const gchar *path,
        gfloat *quality,	/* Quality value from 0.0 to 1.0. */
	gbool *save_as_color
)
{
	gint	width = 320,
		height = -1,
		border_minor = 2;
	GtkAccelGroup *accelgrp = NULL;
	GtkWidget *toplevel = NULL, *client_vbox = NULL;
	GtkAdjustment *quality_adj = NULL;
	GtkWidget	*save_as_greyscale_radio = NULL,
			*save_as_color_radio = NULL;


	/* Reset global got response flag. */
	got_response = FALSE;

	/* Create dialog base widgets. */
	build_std_dialog(
	    width, height,
	    "JPEG Format Options",
	    ref_toplevel,
	    &toplevel, &client_vbox, &accelgrp
	);

	/* Create JPEG format option widgets. */
	if(client_vbox != NULL)
	{
	    gint	scale_width = 150,
			entry_width = 40;
	    GtkWidget *w, *parent = client_vbox, *parent2;
	    GtkEntry *entry;
	    GtkAdjustment *adj;
	    GSList *gslist;


	    /* Quality. */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent2 = w;
	    /* Label. */
	    w = gtk_label_new("Quality:");
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
	    /* Adjustment and hscale. */
	    quality_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
		(quality != NULL) ? (*quality) * 100.0 : 0.0,
		0.0,		/* Minimum. */
		100.0,		/* Maximum. */
		1.0,		/* Step inc. */
		10.0,		/* Page inc. */
		0.0		/* Page size. */
	    );
	    w = gtk_hscale_new(adj);
	    gtk_widget_set_usize(w, scale_width, -1);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	    GUISetWidgetTip(w, SAVE_OPTS_TTM_JPEG_QUALITY);
	    gtk_widget_show(w);
	    /* Entry. */
	    w = gtk_entry_new_with_max_length(80);
	    entry = GTK_ENTRY(w);
	    gtk_widget_set_usize(w, entry_width, -1);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            GUISetWidgetTip(w, SAVE_OPTS_TTM_JPEG_QUALITY);
            gtk_widget_show(w);
	    /* Units label. */
            w = gtk_label_new("%");
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
	    /* Connect value changed signals. */
            gtk_signal_connect(
                GTK_OBJECT(adj), "value_changed",
                GTK_SIGNAL_FUNC(scale_changed_cb),
                (gpointer)entry
            );
            gtk_signal_connect(
                GTK_OBJECT(entry), "changed",
                GTK_SIGNAL_FUNC(entry_changed_adj_cb),
                (gpointer)adj
            );
	    /* Update entry widget value. */
	    gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");

	    /* Save as greyscale or color radios. */
            w = gtk_hbox_new(FALSE, border_minor);
            gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            parent2 = w;
	    /* Greyscale radio. */
	    gslist = NULL;
            save_as_greyscale_radio = w =
		gtk_radio_button_new_with_label(gslist, "GreyScale");
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
	    gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
            /* Color radio. */
            save_as_color_radio = w =
                gtk_radio_button_new_with_label(gslist, "Color");
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
	    /* Set initial radio state. */
	    if((save_as_color != NULL) ? (*save_as_color) : FALSE)
		gtk_toggle_button_set_active(
		    GTK_TOGGLE_BUTTON(save_as_color_radio), TRUE
		);
	    else
                gtk_toggle_button_set_active(
                    GTK_TOGGLE_BUTTON(save_as_greyscale_radio), TRUE
                );


	}


	/* Map toplevel. */
	gtk_widget_show(toplevel);


	/* Enter blocking gtk_main loop. */
	gtk_main();


	/* Get values from widgets if user clicked on save. */
	if(got_response)
	{
	    if((quality_adj != NULL) && (quality != NULL))
		(*quality) = CLIP(
		    quality_adj->value,
		    quality_adj->lower, quality_adj->upper
		) / 100.0;

	    if((save_as_color != NULL) && (save_as_color_radio != NULL))
		(*save_as_color) = GTK_TOGGLE_BUTTON(save_as_color_radio)->active;
	}


	/* Destroy widgets and related resources. */
	gtk_window_set_modal(GTK_WINDOW(toplevel), FALSE);
	gtk_window_set_transient_for(GTK_WINDOW(toplevel), NULL);
	gtk_widget_destroy(toplevel);
	gtk_accel_group_unref(accelgrp);


	/* Return user response. */
	return(got_response);
}

/*
 *	PNG format options querying.
 *
 *      Blocks execution untill response is recieved.
 *      Returns TRUE if values were fetched or FALSE if user canceled.
 */
gint iv_query_png_opts(
        GtkWidget *ref_toplevel,
        const gchar *path,
	gint *compression,	/* Value from 0 to 9. */
        gint *format,           /* 0 = B&W  1 = Greyscale
                                 * 2 = Greyscale Alpha
                                 * 3 = RGB  4 = RGBA
                                 */
	gbool *interlaced
)
{
        gint    width = 320,
                height = -1,
                border_minor = 2,
		border_major = 5;
        GtkAccelGroup *accelgrp = NULL;
        GtkWidget *toplevel = NULL, *client_vbox = NULL;
        GtkAdjustment *compression_adj = NULL;
        GtkWidget       *format_greyscale_radio = NULL,
                        *format_rgb_radio = NULL,
			*alpha_toggle = NULL;
	GtkWidget *interlaced_toggle = NULL;


        /* Reset global got response flag. */
        got_response = FALSE;

        /* Create dialog base widgets. */
        build_std_dialog(
            width, height,
            "PNG Format Options",
            ref_toplevel,
            &toplevel, &client_vbox, &accelgrp
        );

        /* Create PNG format option widgets. */
        if(client_vbox != NULL)
        {
            gint        scale_width = 150,
                        entry_width = 40;
            GtkWidget *w, *parent = client_vbox, *parent2, *parent3;
            GtkEntry *entry;
            GtkAdjustment *adj;
            GSList *gslist;


            /* Compression. */
            w = gtk_hbox_new(FALSE, border_minor);
            gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            parent2 = w;
            /* Label. */
            w = gtk_label_new("Compression:");
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            /* Adjustment and hscale. */
            compression_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
		(gfloat)((compression != NULL) ? *compression : 0),
                0.0,		/* Minimum. */
                9.0,		/* Maximum. */
                1.0,            /* Step inc. */
                1.0,		/* Page inc. */
                0.0		/* Page size. */
            );
            w = gtk_hscale_new(adj);
            gtk_widget_set_usize(w, scale_width, -1);
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
            GUISetWidgetTip(w, SAVE_OPTS_TTM_PNG_COMPRESSION);
            gtk_widget_show(w);
            /* Entry. */
            w = gtk_entry_new_with_max_length(80);
            entry = GTK_ENTRY(w);
            gtk_widget_set_usize(w, entry_width, -1);
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            GUISetWidgetTip(w, SAVE_OPTS_TTM_PNG_COMPRESSION);
            gtk_widget_show(w);
            /* Connect value changed signals. */
            gtk_signal_connect(
                GTK_OBJECT(adj), "value_changed",
                GTK_SIGNAL_FUNC(scale_changed_cb),
                (gpointer)entry
            );
            gtk_signal_connect(
                GTK_OBJECT(entry), "changed",
                GTK_SIGNAL_FUNC(entry_changed_adj_cb),
                (gpointer)adj
            );
            /* Update entry widget value. */
            gtk_signal_emit_by_name(GTK_OBJECT(adj), "value_changed");


	    /* Hbox to hold format and alpha widgets in two separate sub
	     * hboxes.
	     */
            w = gtk_hbox_new(FALSE, border_major);
            gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            parent2 = w;

	    /* Hbox for format greyscale and rgb radios. */
	    w = gtk_hbox_new(FALSE, border_minor);
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            parent3 = w;
	    /* Format greyscale and rgb radios. */
            /* Greyscale radio. */
            gslist = NULL;
            format_greyscale_radio = w =
                gtk_radio_button_new_with_label(gslist, "GreyScale");
            gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
            /* RGB radio. */
            format_rgb_radio = w =
                gtk_radio_button_new_with_label(gslist, "RGB");
            gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            /* Set initial radio state. */
	    switch((format != NULL) ? *format : 0)
	    {
	      /* B&W, Greyscale, or Greyscale Alpha. */
	      case 0: case 1: case 2:
		gtk_toggle_button_set_active(
                    GTK_TOGGLE_BUTTON(format_greyscale_radio), TRUE
                );
		break;
	      /* RGB or RGBA. */
	      case 3: case 4:
                gtk_toggle_button_set_active(
                    GTK_TOGGLE_BUTTON(format_rgb_radio), TRUE
                );
		break;
	    }

	    /* Hbox for alpha channel check. */
            w = gtk_hbox_new(FALSE, border_minor);
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            parent3 = w;
	    /* Alpha channel check. */
            alpha_toggle = w = gtk_check_button_new_with_label("Alpha");
            gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    /* Set initial state. */
	    switch((format != NULL) ? *format : 0)
            {
	      /* Greyscale Alpha or RGBA. */
	      case 2: case 4:
                GTK_TOGGLE_BUTTON(w)->active = TRUE;
		break;
	    }
            GUISetWidgetTip(w, SAVE_OPTS_TTM_PNG_ALPHA);
            gtk_widget_show(w);


	    /* Interlaced. */
            w = gtk_hbox_new(FALSE, border_minor);
            gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            parent2 = w;
	    /* Check button. */
	    interlaced_toggle = w = gtk_check_button_new_with_label("Interlaced");
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    if((interlaced != NULL) ? *interlaced : 0)
		GTK_TOGGLE_BUTTON(w)->active = TRUE;
            GUISetWidgetTip(w, SAVE_OPTS_TTM_PNG_INTERLACED);
            gtk_widget_show(w);
	}


        /* Map toplevel. */
        gtk_widget_show(toplevel);


        /* Enter blocking gtk_main loop. */
        gtk_main();


        /* Get values from widgets if user clicked on save. */
        if(got_response)
        {
            if((compression_adj != NULL) && (compression != NULL))
                (*compression) = (gint)CLIP(
		    compression_adj->value,
		    compression_adj->lower, compression_adj->upper
		);
	    if((format != NULL) && (format_rgb_radio != NULL) &&
	       (alpha_toggle != NULL)
	    )
	    {
		/* RGB or RGBA? */
		if(GTK_TOGGLE_BUTTON(format_rgb_radio)->active)
		{
		    /* RGBA? */
		    if(GTK_TOGGLE_BUTTON(alpha_toggle)->active)
			(*format) = 4;
		    else
			(*format) = 3;
		}
		/* Greyscale or Greyscale Alpha. */
		else
		{
		    /* Greyscale Alpha? */
                    if(GTK_TOGGLE_BUTTON(alpha_toggle)->active)
			(*format) = 2;
                    else
			(*format) = 1;
		}
	    }
	    if((interlaced != NULL) && (interlaced_toggle != NULL))
		(*interlaced) = GTK_TOGGLE_BUTTON(interlaced_toggle)->active;
        }


        /* Destroy widgets and related resources. */
        gtk_window_set_modal(GTK_WINDOW(toplevel), FALSE);
        gtk_window_set_transient_for(GTK_WINDOW(toplevel), NULL);
        gtk_widget_destroy(toplevel);
        gtk_accel_group_unref(accelgrp);


        /* Return user response. */
        return(got_response);
}

/*
 *	TGA format options querying.
 *
 *      Blocks execution untill response is recieved.
 *      Returns TRUE if values were fetched or FALSE if user canceled.
 */
gint iv_query_tga_opts(
        GtkWidget *ref_toplevel,
        const gchar *path,
        gint *format            /* 0 Greyscale
                                 * 1 = RGB
                                 * 2 = RGBA
                                 */
)
{
        gint    width = 320,
                height = -1,
                border_minor = 2,
                border_major = 5;
        GtkAccelGroup *accelgrp = NULL;
        GtkWidget *toplevel = NULL, *client_vbox = NULL;
        GtkWidget       *format_greyscale_radio = NULL,
                        *format_rgb_radio = NULL,
                        *format_rgba_radio = NULL;


        /* Reset global got response flag. */
        got_response = FALSE;

        /* Create dialog base widgets. */
        build_std_dialog(
            width, height,
            "TGA Format Options",
            ref_toplevel,
            &toplevel, &client_vbox, &accelgrp
        );

        /* Create TGA format option widgets. */
        if(client_vbox != NULL)
        {
            GtkWidget *w, *parent = client_vbox, *parent2, *parent3;
            GSList *gslist;


            w = gtk_hbox_new(FALSE, border_major);
            gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            parent2 = w;

            /* Hbox for format greyscale, rgb, and rgba radios. */
            w = gtk_hbox_new(FALSE, border_minor);
            gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            parent3 = w;
            /* Greyscale radio */
            gslist = NULL;
            format_greyscale_radio = w =
                gtk_radio_button_new_with_label(gslist, "GreyScale");
            gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
            /* RGB radio */
            format_rgb_radio = w =
                gtk_radio_button_new_with_label(gslist, "RGB");
            gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
            gtk_widget_show(w);
            gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	    /* RGBA radio */
            format_rgba_radio = w =
                gtk_radio_button_new_with_label(gslist, "RGBA");
            gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
            gtk_widget_show(w);

            /* Set initial radio state. */
            switch((format != NULL) ? *format : 0)
            {
              case 2:
                gtk_toggle_button_set_active(
                    GTK_TOGGLE_BUTTON(format_rgba_radio), TRUE
                );
                break;

              case 1:
                gtk_toggle_button_set_active(
                    GTK_TOGGLE_BUTTON(format_rgb_radio), TRUE
                );
                break;

              default:
                gtk_toggle_button_set_active(
		    GTK_TOGGLE_BUTTON(format_greyscale_radio), TRUE
		);
		break;
	    }
	}


        /* Map toplevel. */
        gtk_widget_show(toplevel);


        /* Enter blocking gtk_main loop. */
        gtk_main();


        /* Get values from widgets if user clicked on save. */
        if(got_response)
        {
	    if(format != NULL)
	    {
                if(GTK_TOGGLE_BUTTON(format_rgba_radio)->active)
		    *format = 2;
		else if(GTK_TOGGLE_BUTTON(format_rgb_radio)->active)
                    *format = 1;
                else
                    *format = 0;
	    }
	}

        /* Destroy widgets and related resources. */
        gtk_window_set_modal(GTK_WINDOW(toplevel), FALSE);
        gtk_window_set_transient_for(GTK_WINDOW(toplevel), NULL);
        gtk_widget_destroy(toplevel);
        gtk_accel_group_unref(accelgrp);


        /* Return user response. */
        return(got_response);
}
